# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 16:12:40 2023

@author: 20225533
"""

# %%
###############################################################################
# IMPORT LIBRARIES
###############################################################################
# Code developed by Ilaria Fichera for the analysis of the FVM method adapted solving the 3D diffusion equation with one intermittent omnidirectional sound source
# Import modules
import itertools
import json
import os
import pickle
import time
import types
import pandas as pd

from math import ceil
from math import log

import gmsh

from Diffusion_Module.FiniteVolumeMethod.FunctionRT import t60_decay
from Diffusion_Module.FiniteVolumeMethod.FunctionClarity import *
from Diffusion_Module.FiniteVolumeMethod.FunctionDefinition import *
from Diffusion_Module.FiniteVolumeMethod.FunctionCentreTime import *
from Diffusion_Module.FiniteVolumeMethod.CreateMeshFVM import generate_mesh

# Silvin: debugging
import logging
import traceback

logger = logging.getLogger(__name__)

# %%
###############################################################################
# SURFACE MATERIALS FUNCTIONS
###############################################################################


def de_method(json_file_path=None):

    if not gmsh.is_initialized():
        gmsh.initialize()

    st = time.time()  # start time of calculation
    result_container = {}
    if json_file_path is not None:
        with open(json_file_path, "r") as json_file:
            result_container = json.load(json_file)

    # Checking whether the 'should_cancel' flag has been set to True by the user
    # Do not call this function all the time, as it is quite heavy
    # This function should be called in the main calculation loop
    def check_should_cancel(json_file_path_in):
        try:
            if json_file_path_in is not None:
                with open(json_file_path_in, "r") as json_file_to_check:
                    data = json.load(json_file_to_check)

            # Update the specified field value
            if "should_cancel" in data:
                return data["should_cancel"]

        except Exception as e:
            print("check_should_cancel returned: " + str(e))
            print(traceback.format_exc())

    if check_should_cancel(json_file_path):
        return

    # %%
    ###############################################################################
    # INPUT VARIABLES
    ###############################################################################

    # Source position
    coord_source = [
        0.5,
        0.5,
        1.0,
    ]  # coordinates of the source position in an list: x , y and z direction in meters [m]

    # Receiver position
    coord_rec = [
        2.0,
        0.5,
        1.0,
    ]  # coordinates of the receiver position in an list: x , y and z direction in meters [m]

    # Type of Calculation
    # Choose "decay" if the objective is to calculate the energy decay of the room with all its energetic parameters;
    # Choose "stationarysource" if the aim is to understand the behaviour of a room subject to a stationary source
    tcalc = "decay"

    # Frequency range
    fc_low = 125
    fc_high = 2000
    num_octave = 1

    x_frequencies = num_octave * log(fc_high / fc_low) / log(2)
    nBands = int(num_octave * log(fc_high / fc_low) / log(2) + 1)
    center_freq = fc_low * np.power(2, ((np.arange(0, x_frequencies + 1) / num_octave)))

    # Time discretization
    dt = 1 / 20000  # time discretization

    # Air absorption coefficient
    m_atm = 0  # air absorption coefficient [1/m]

    # %%
    ###############################################################################
    # FIXED INPUTS
    ###############################################################################
    # General settings

    # Set initial condition - Source Info (interrupted method)
    Ws = 0.01  # Source point power [Watts] interrupted after "sourceon_time" seconds; 10^-2 W => correspondent to 100dB

    # Absorption term and Absorption coefficients
    th = 3  # int(input("Enter type Absortion conditions (option 1,2,3):"))
    # options Sabine (th=1), Eyring (th=2) and modified by Xiang (th=3)

    # Reference pressure in Pa
    pRef = 2 * (10**-5)  # Reference pressure in Pa

    # Air density
    rho = 1.21  # air density [kg.m^-3] at 20°C

    # %%
    ###############################################################################
    # INITIALISE GMSH
    ###############################################################################
    msh_file_path = "C:\\Users\\20225521\\Desktop\\Room_acoustics\\ui_backend\\Diffusion\\FiniteVolumeMethod\\3x3x3.msh"  # Insert file name, msh file created from sketchUp and then gmsh

    # gmsh.fltk.run() #run the file to see it in gmsh
    dim = (
        -1
    )  # dimensions of the entities, 0 for points, 1 for curves/edge/lines, 2 for surfaces, 3 for volumes, -1 for all the entities
    tag = -1  # all the nodes of the room

    # %%
    ###############################################################################
    # Integraiton changes added by Hassan: ;) Do not modify any of these please!
    ###############################################################################
    use_default = True

    if result_container:
        simulation_settings = result_container["simulationSettings"]
        use_default = result_container["settingsPreset"] == "Default"

        if use_default:
            c0 = 343
        else:
            c0 = simulation_settings["de_c0"]

        coord_source = [
            result_container["results"][0]["sourceX"],
            result_container["results"][0]["sourceY"],
            result_container["results"][0]["sourceZ"],
        ]

        coord_rec = [
            result_container["results"][0]["responses"][0]["x"],
            result_container["results"][0]["responses"][0]["y"],
            result_container["results"][0]["responses"][0]["z"],
        ]
        geo_file_path = result_container["geo_path"]
        msh_file_path = result_container["msh_path"]
        generate_mesh(
            geo_file_path, msh_file_path, 1
        )  # TODO: make this dependent on the room dimensions. We don't need an lc of 1 meter at all times..
    else:
        c0 = 343  # adiabatic speed of sound [m.s^-1]

    mesh = gmsh.open(msh_file_path)  # open the file

    # Absorption term for boundary conditions
    def abs_term(th, abscoeff_list):
        Absx_array = np.array([])
        for abs_coeff in abscoeff_list:
            # print(abs_coeff)
            if th == 1:
                Absx = (c0 * abs_coeff) / 4  # Sabine
            elif th == 2:
                Absx = (c0 * (-log(1 - abs_coeff))) / 4  # Eyring
            elif th == 3:
                Absx = (c0 * abs_coeff) / (2 * (2 - abs_coeff))  # Modified by Xiang
            Absx_array = np.append(Absx_array, Absx)
        return Absx_array

    def surface_materials():
        vGroups = gmsh.model.getPhysicalGroups(
            -1
        )  # these are the entity tag and physical groups in the msh file.
        vGroupsNames = (
            []
        )  # these are the entity tag and physical groups in the msh file + their names
        for iGroup in vGroups:
            dimGroup = iGroup[
                0
            ]  # entity tag: 1 lines, 2 surfaces, 3 volumes (1D, 2D or 3D)
            tagGroup = iGroup[
                1
            ]  # physical tag group (depending on material properties defined in SketchUp)
            namGroup = gmsh.model.getPhysicalName(
                dimGroup, tagGroup
            )  # names of the physical groups defined in SketchUp
            alist = [
                dimGroup,
                tagGroup,
                namGroup,
            ]  # creates a list of the entity tag, physical tag group and name
            # print(alist)
            vGroupsNames.append(alist)

        # Initialize a list to store surface tags and their absorption coefficients
        surface_absorption = (
            []
        )  # initialization absorption term (alpha*surfaceofwall) for each wall of the room
        triangle_face_absorption = (
            []
        )  # initialization absorption term for each triangle face at the boundary and per each wall
        absorption_coefficient = {}

        for group in vGroupsNames:
            if group[0] != 2:
                continue
            name_group = group[2]
            name_split = name_group.split("$")
            name_abs_coeff = name_split[0]
            if result_container:
                abscoeff = result_container["absorption_coefficients"][name_abs_coeff]
            else:
                abscoeff = input(
                    f"Enter absorption coefficient for frequency {fc_low} to {fc_high} for {name_abs_coeff}:"
                )

            abscoeff = abscoeff.split(",")
            # abscoeff = [float(i) for i in abscoeff][-1] #for one frequency

            if result_container:
                if use_default:
                    abscoeff_list = [
                        float(i) for i in abscoeff
                    ]  # for multiple frequencies
                else:
                    if simulation_settings["de_absorption_override"] == "yes":
                        abscoeff_list = [1 - simulation_settings["de_R"] ** 2] * len(
                            abscoeff
                        )
                    else:
                        abscoeff_list = [
                            float(i) for i in abscoeff
                        ]  # for multiple frequencies

            physical_tag = group[1]  # Get the physical group tag
            entities = gmsh.model.getEntitiesForPhysicalGroup(
                2, physical_tag
            )  # Retrieve all the entities in this physical group (the entities are the number of walls in the physical group)

            Abs_term = abs_term(
                th, abscoeff_list
            )  # calculates the absorption term based on the type of boundary condition th
            for entity in entities:
                absorption_coefficient[entity] = abscoeff_list
                surface_absorption.append(
                    (entity, Abs_term)
                )  # absorption term (alpha*surfaceofwall) for each wall of the room
                surface_absorption = sorted(surface_absorption, key=lambda x: x[0])

        for entity, Abs_term in surface_absorption:
            triangle_faces, _ = gmsh.model.mesh.getElementsByType(
                2, entity
            )  # Get all the triangle faces for the current surface
            triangle_face_absorption.extend(
                [Abs_term] * len(triangle_faces)
            )  # Append the Abs_term value for each triangle face

        return absorption_coefficient, surface_absorption, triangle_face_absorption

    # FUNCTION CALLED HERE
    absorption_coefficient, surface_absorption, triangle_face_absorption = (
        surface_materials()
    )
    print(
        "Correctly inputted surface materials. Starting initial geometry calculations..."
    )

    # %%
    ###############################################################################
    # GMSH GET NODES, VOLUME ELEMENTS AND BOUNDARY ELEMENTS
    ###############################################################################
    def get_nodes_elem():
        # Nodes
        tag = -1  # all the nodes of the room
        nodeTags, coords, parametricCoord = gmsh.model.mesh.getNodes(
            dim, tag
        )  # gets the tags for each node and the coordinates of each node
        nodecoords = coords.reshape(
            (-1, 3)
        )  # coordinates reshaped in a matrix 3xnumber of nodes

        node_indices = {tag: index for (index, tag) in enumerate(nodeTags)}

        # Element Types
        elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(dim, tag)
        # elemTypes = 1 for lines, 2 for surfaces, 4 for tetrahedron
        # elemTags =  list of list of lines, boundary elements (surfaces) and volume elements (tetrahedron)
        # elemNodeTags = did not understand this yet, probably a tag gives to each line, surface and tetrahedron
        for e_type in elemTypes:
            if (
                e_type == 1
            ):  # if the e_type = 1, then get all the elements (lines) with that e_type
                tag = -1
                edgeEl, edgeNodeTagstype = gmsh.model.mesh.getElementsByType(
                    e_type, tag
                )  # get the edge element tags and the nodes; edgeEl are lines
                # edgeEl = numbered edge elements (lines)
            elif (
                e_type == 2
            ):  # if the e_type = 2, then get all the elements (surfaces) with that e_type
                tag = -1
                bounEl, bounNodeTagstype = gmsh.model.mesh.getElementsByType(
                    e_type, tag
                )  # get the boundary/surface element tags and the nodes; bounEl are surfaces
                bounNode = bounNodeTagstype.reshape((-1, 3))
                # boundEl = numbered boundary elements (surfaces)
                # bounNode = nodes of the surfaces
            elif e_type == 4:
                tag = (
                    -1
                )  # if the e_type = 4, then get all the elements (tetrahedron) with that e_type
                voluEl, voluNodeTagstype = gmsh.model.mesh.getElementsByType(
                    e_type, tag
                )  # get the volume element tags and the nodes; voluEl are tetrahedrons
                voluNode = voluNodeTagstype.reshape((-1, 4))
                # voluEl = numbered volume elements (tetrahedron)
                # voluNode = nodes of the tetrahedon
            else:
                print("The procedure is not possible")

        velemNodes = elemNodeTags[2].reshape((-1, 4))  # nodes per each tetrahedron
        belemNodes = elemNodeTags[1].reshape((-1, 3))  # nodes per each surface boundary

        eelement = 0
        for elem in range(len(edgeEl)):
            eelement = eelement + 1  # scalar number of the edge elements

        # Volume Element dictionary + nodes of each volume elements (4 nodes per element)
        volumeEl_dict = {}  # Initialization Dictionary of volumelements + its nodes
        for i in range(len(voluEl)):
            # print(i)
            volumeEl_dict[voluEl[i]] = velemNodes[
                i
            ]  # Dictionary of volumelements + its nodes

        # Boundary Element dictionary + node per each surface elements (3 nodes per element)
        boundaryEl_dict = (
            {}
        )  # Initialization Dictionary of boundary elements + its nodes
        for i in range(len(bounEl)):
            boundaryEl_dict[bounEl[i]] = belemNodes[
                i
            ]  # Dictionary of boundary elements + its nodes

        return (
            nodecoords,
            node_indices,
            bounEl,
            bounNode,
            voluEl,
            voluNode,
            belemNodes,
            velemNodes,
            boundaryEl_dict,
            volumeEl_dict,
        )

    # FUNCTION CALLED HERE
    (
        nodecoords,
        node_indices,
        bounEl,
        bounNode,
        voluEl,
        voluNode,
        belemNodes,
        velemNodes,
        boundaryEl_dict,
        volumeEl_dict,
    ) = get_nodes_elem()

    # %%
    ###############################################################################
    # CALCULATION OF VOLUME CELLS AND CENTRE OF VOLUME
    ###############################################################################

    def velem_volume_centre():
        # Calculation of volume cells and centre of volume
        vcell_dict = {}  # volume of each element tetrahedron initialization
        centre_cell = {}  # centre of the element tetrahedron initialization
        for i in volumeEl_dict.keys():
            coord_centre_cell = np.zeros(3)
            centre_cell[i] = []
            # print(i)
            vc0 = nodecoords[
                node_indices[volumeEl_dict[i][0]], :
            ]  # coordinates of node 0
            # vc0 = gmsh.model.mesh.getNode(volumeEl_dict[i][0])[0] #Coordinates of the node number zero of the volume element i
            # print(nc0)
            vc1 = nodecoords[node_indices[volumeEl_dict[i][1]], :]
            vc2 = nodecoords[node_indices[volumeEl_dict[i][2]], :]
            vc3 = nodecoords[node_indices[volumeEl_dict[i][3]], :]
            # vc1 = gmsh.model.mesh.getNode(volumeEl_dict[i][1])[0] #Coordinates of the node number one of the volume element i
            # print(nc1)
            # vc2 = gmsh.model.mesh.getNode(volumeEl_dict[i][2])[0] #Coordinates of the node number two of the volume element i
            # print(nc2)
            # vc3 = gmsh.model.mesh.getNode(volumeEl_dict[i][3])[0] #Coordinates of the node number three of the volume element i
            # print(nc3)
            for j in range(3):  # three coordinates per each node
                coord_centre_cell[j] = (
                    vc0[j] + vc1[j] + vc2[j] + vc3[j]
                ) / 4  # coordinates of the centre of each volume element
                centre_cell[i].append(coord_centre_cell[j])
            vcell_dict[i] = (
                abs(np.dot(np.cross(vc1 - vc3, vc2 - vc3), vc0 - vc3)) / 6
            )  # volume of each volume element

        # The dictionary centre cell is modified into an array of floats (list)
        cell_center = np.array(
            ()
        )  # initialization of an array of centre cell coordinates from the centre_cell dictionary
        for key in centre_cell:
            cell_center = np.append(cell_center, centre_cell[key])
        cell_center = cell_center.reshape((-1, 3))

        cell_volume = np.array(
            ()
        )  # initialization of an array of cell volumes from the vcell_dict dictionary
        for key in vcell_dict:
            cell_volume = np.append(cell_volume, vcell_dict[key])

        return cell_center, cell_volume

    # FUNCTION CALLED HERE
    cell_center, cell_volume = velem_volume_centre()

    # %%
    ###############################################################################
    # CALCULATION OF BOUNDARY ELEMENTS AND CENTRE OF AREA (might not need this function actually)
    ###############################################################################
    # Calculation of boundary elements area and centre
    def belem_area_centre():
        barea_dict = {}  # surface of each element boundary initialization
        centre_area = {}  # centre of the element tetrahedron initialization
        for i in boundaryEl_dict.keys():
            coord_centre_area = np.zeros(3)
            centre_area[i] = []
            # print(i)
            bc0 = nodecoords[node_indices[boundaryEl_dict[i][0]], :]

            # bc0 = gmsh.model.mesh.getNode(boundaryEl_dict[i][0])[0]
            # bnodeCoord_dict[boundaryEl_dict[i][0]] #Coordinates of the node number zero of the volume element i
            # print(nc0)
            bc1 = nodecoords[node_indices[boundaryEl_dict[i][1]], :]
            # gmsh.model.mesh.getNode(boundaryEl_dict[i][1])[0] #Coordinates of the node number one of the volume element i
            # print(nc1)
            bc2 = nodecoords[node_indices[boundaryEl_dict[i][2]], :]
            # gmsh.model.mesh.getNode(boundaryEl_dict[i][2])[0] #Coordinates of the node number two of the volume element i
            # print(nc2)
            for j in range(3):
                coord_centre_area[j] = (
                    bc0[j] + bc1[j] + bc2[j]
                ) / 3  # coordinates of the centre of each volume element
                centre_area[i].append(coord_centre_area[j])
            barea_dict[i] = (
                abs(sum(np.cross(bc2 - bc1, bc1 - bc0))) / 2
            )  # volume of each volume element

        return barea_dict, centre_area

    # FUNCTION CALLED HERE
    barea_dict, centre_area = belem_area_centre()

    # %%
    ###############################################################################
    # CALCULATION OF NEIGHBOURS (might not need this function actually)
    ###############################################################################
    # Neighbours calculation; What are the neighbours faces of each volume? 3 per each minimum?
    def get_neighbour_faces():
        facenodes = gmsh.model.mesh.getElementFaceNodes(
            4, 3
        )  # 4 is the element type (tetrahedron) and three are the nodes per each face #get all the face tags of all the faces of the tetrahedrons

        # Computing face x tetrahedon incidence
        faces = []  # initialization of list of tuples of the faces nodes
        fxt = (
            {}
        )  # dictionary with keys as the nodes of each face and values the volume elements of which this face is neighbour
        for i in range(
            0, len(facenodes), 3
        ):  # per each element basically, goes trhough the nodes of each face 3by3
            # print(i)
            f = tuple(
                sorted(facenodes[i : i + 3])
            )  # nodes of each face put in a tuple from node i to node i plus 3
            faces.append(f)
            tet = voluEl[
                i // 12
            ]  # volume element number at which the faces are associated?
            if (
                not f in fxt
            ):  # if the face f (with its node) is already in the dictionary, just append the volume element neighbour to
                fxt[f] = [tet]
            else:
                fxt[f].append(tet)

        # Computing neighbors by face
        txt = (
            {}
        )  # dictionary with keys as the tetrahedron tag and values as the tet that are neighbours (the tet at the boundary are not counted)
        for i in range(0, len(faces)):
            # print(i)
            f = faces[i]  # f is a tuple of the nodes of the face into consideration
            tet = voluEl[i // 4]  # tetrahedron at which the face is neighbour
            if (
                not tet in txt
            ):  # if the tet is not in the dictionary, add it, otherwise append the new tt
                txt[tet] = []
            for tt in fxt[f]:
                if tt != tet:
                    txt[tet].append(
                        int(tt - (voluEl[0] - 1))
                    )  # volumes neighbours to each volume
        for values in txt.values():
            while (
                len(values) < 4
            ):  # if there are less than 4 tetrahedron neighbours it means that the others are boundary tetrahedrons, therefore add a zero for each tetrahedron missing
                values.append(0)

        neighbourVolume = np.array(
            ()
        )  # initialization of an array with the neighbours tetrahedron per each tetrahedron in order from 0 to the number of tetrahedrons
        for key in txt:
            neighbourVolume = np.append(neighbourVolume, txt[key])
        for item in neighbourVolume:
            item = int(item)

        neighbourVolume = neighbourVolume.reshape(
            (-1, 4)
        )  # reshape the array so that we have the 4 tetrahedrons neighbours of the tetrahedron in consideration

        return fxt, txt, neighbourVolume

    # FUNCTION CALLED HERE
    if check_should_cancel(json_file_path):
        return
    fxt, txt, neighbourVolume = get_neighbour_faces()
    print(
        "Completed initial geometry calculation. Starting internal tetrahedrons calculations..."
    )

    # %%
    ###############################################################################
    # CALCULATION OF INTERIOR TETRAHEDRONS
    ###############################################################################
    # Interior Tetrahedrons calculations

    def interior_tetra():
        interior_tet = np.zeros(
            (len(voluEl), len(voluEl))
        )  # initialization matrix of tetrahedron per tetrahedron

        for i in range(len(voluEl)):  # for each tetrahedron, take its centre
            # print(i)
            cell_center_i = cell_center[i]
            for j in range(len(voluEl)):  # for each tetrahedron, take its centre
                # cell_center_j = cell_center[j]
                # print(j)
                if (
                    i != j
                ):  # if the tetrahedrons are not the same one, then check if there are shared nodes in between the two tetrahedron i and j
                    shared_nodes = np.intersect1d(velemNodes[i], velemNodes[j])
                    # shared_nodes = []
                    # count = 0
                    # for node in velemNodes[i]: #for each node in tetrahedron i
                    #    print(node)
                    #    if node in velemNodes[j]: #if each node of the tetrahedron i is in nodelist of tetrahedron j
                    #        count += 1
                    #        shared_nodes.append(node) #append the node that it is in common
                    if (
                        len(shared_nodes) == 3
                    ):  # after have done this for all the nodes, if the cound is 3 then calculate the shared area between the tetrahedrons
                        sc0 = nodecoords[node_indices[shared_nodes[0]], :]
                        # sc0 = gmsh.model.mesh.getNode(shared_nodes[0])[0] #coordinates of node 0
                        sc1 = nodecoords[node_indices[shared_nodes[1]], :]
                        # sc1 = gmsh.model.mesh.getNode(shared_nodes[1])[0] #coordinates of node 1
                        sc2 = nodecoords[node_indices[shared_nodes[2]], :]
                        # sc2 = gmsh.model.mesh.getNode(shared_nodes[2])[0] #coordinates of node 2
                        shared_area = (
                            np.linalg.norm(np.cross(sc2 - sc0, sc1 - sc0)) / 2
                        )  # compute shared area
                        shared_distance = np.linalg.norm(cell_center_i - cell_center[j])
                        # sqrt((abs(cell_center_i[0] - cell_center_j[0]))**2 + (abs(cell_center_i[1] - cell_center_j[1]))**2 + (abs(cell_center_i[2] - cell_center_j[2]))**2) #distance between volume elements
                        interior_tet[i, j] = (
                            shared_area / shared_distance
                        )  # division between shared area and shared distance
                    else:
                        shared_area = 0
                        interior_tet[i, j] = shared_area

        interior_tet_sum = np.sum(
            interior_tet, axis=1
        )  # sum of interior_tet per columns (so per i element)

        return interior_tet, interior_tet_sum

    # FUNCTION CALLED HERE
    if check_should_cancel(json_file_path):
        return
    interior_tet, interior_tet_sum = interior_tetra()
    print(
        "Completed internal tetrahedrons calculation. Starting boundary tetrahedrons calculations..."
    )

    # %%
    ###############################################################################
    # CALCULATION OF ENTIRE SURFACE AREA PER MATERIAL
    ###############################################################################
    # Calculation surface areas
    def surface_area():
        surface_areas = {}
        for entity, Abs_term in surface_absorption:
            face_nodes_per_entity = gmsh.model.mesh.getElementFaceNodes(
                2, 3, tag=entity
            )
            surf_area_tot = 0
            for i in range(
                0, len(face_nodes_per_entity), 3
            ):  # per each element basically, goes trhough the nodes of each face 3by3
                # print(i)
                f = tuple(sorted(face_nodes_per_entity[i : i + 3]))
                fc0 = nodecoords[node_indices[f[0]], :]
                # fc0 = gmsh.model.mesh.getNode(f[0])[0] #coordinates of vertix 0
                fc1 = nodecoords[node_indices[f[1]], :]
                fc2 = nodecoords[node_indices[f[2]], :]
                # fc1 = gmsh.model.mesh.getNode(f[1])[0] #coordinates of vertix 1
                # fc2 = gmsh.model.mesh.getNode(f[2])[0] #coordinates of vertix 2
                face_area = 0.5 * np.linalg.norm(
                    np.cross(fc1 - fc0, fc2 - fc0)
                )  # Compute the area using half of the cross product's magnitude
                surf_area_tot += face_area
                surface_areas[entity] = surf_area_tot
        return surface_areas

    # FUNCTION CALLED HERE
    surface_areas = surface_area()

    # %%
    ###############################################################################
    # CALCULATION OF BOUNDARY ELEMENTS
    ###############################################################################
    # FACE AREA & boundary_areas
    def boundary_triang():
        total_boundArea = 0  # initialization of total surface area of the room
        boundary_areas = (
            []
        )  # Initialize a list to store boundary_areas values for each tetrahedron
        import itertools

        face_areas = np.zeros(
            len(velemNodes)
        )  # Per each tetrahedron, if there is a face that is on the boundary, include the area, otehrwise zero
        for idx, element in enumerate(
            velemNodes
        ):  # for index and element in the number of tetrahedrons
            # if idx == 491:
            tetrahedron_boundary_areas = (
                0  # initialization tetrahedron face on boundary*its absorption term
            )
            total_tetrahedron_boundary_areas = np.zeros(
                nBands
            )  # initialization total tetrahedron face on boundary*its absorption term if there are more than one face in the tetrahedron that is on the boundary
            # print(idx)
            node_combinations = [
                list(nodes) for nodes in itertools.combinations(element, 3)
            ]  # all possible combinations of the nodes of the tetrahedrons (it checks also for the order of the nodes in the same combination)
            # Check if the nodes are in any order in bounNode
            is_boundary = False  # variable to say that at the beginning the face in not on a boundary
            for nodes in node_combinations:  # for each node in each combination
                for surface_idx, surface in enumerate(
                    bounNode
                ):  # for index and surface in the number of nodes
                    surface_set = sorted(
                        set(surface)
                    )  # creates a set of the surface nodes
                    surface_set_idx = surface_idx
                    nodes_set = sorted(
                        set(nodes)
                    )  # create a set of the node combination of the tetrahedron into consideration
                    # surface_list = list(surface)
                    if (
                        nodes_set == surface_set
                    ):  # if these are equal, it means that the tetrahedron into consideration has a surface in the boundary and therefore is_boundary gets the value of True.
                        # print(surface_set)
                        # print(surface_list)
                        is_boundary = True
                        if (
                            is_boundary
                        ):  # if the surface is at the boundary, then take the coordinates of each vertix
                            # Convert the vertices to NumPy arrays for vector operations
                            bc0 = nodecoords[node_indices[nodes[0]], :]
                            bc1 = nodecoords[node_indices[nodes[1]], :]
                            bc2 = nodecoords[node_indices[nodes[2]], :]

                            # bc0 = gmsh.model.mesh.getNode(nodes[0])[0] #coordinates of vertix 0
                            # bc1 = gmsh.model.mesh.getNode(nodes[1])[0] #coordinates of vertix 1
                            # bc2 = gmsh.model.mesh.getNode(nodes[2])[0] #coordinates of vertix 2

                            face_area = 0.5 * np.linalg.norm(
                                np.cross(bc1 - bc0, bc2 - bc0)
                            )  # Compute the area using half of the cross product's magnitude
                            # print(face_area)

                            face_areas[idx] = (
                                face_area  # area of the surface that is on boundary per each tetrahedron
                            )
                            total_boundArea += (
                                face_area  # add to the total boundary area
                            )

                            if face_area > 0:
                                # Use the index to access the corresponding absorption area
                                face_absorption_product = (
                                    face_area
                                    * triangle_face_absorption[surface_set_idx]
                                )  # calculate the product between the area*the correspondent absorption term
                                # print(face_absorption_product)

                                tetrahedron_boundary_areas += face_absorption_product  # add the calculation to the tetrahedron correspondent

                                total_tetrahedron_boundary_areas = tetrahedron_boundary_areas  # if there are multiple surfaces on the boundary per each tetrahedron, then add also the second and the third one

            boundary_areas.append(
                np.array(total_tetrahedron_boundary_areas)
            )  # Append the total boundary_areas for the tetrahedron to the list
        # print(total_tetrahedron_boundary_areas)
        boundary_areas = np.array(boundary_areas)
        boundary_areas = boundary_areas.T
        return boundary_areas, total_boundArea

    # FUNCTION CALLED HERE
    if check_should_cancel(json_file_path):
        return
    boundary_areas, total_boundArea = boundary_triang()
    print(
        "Completed boundary tetrahedrons calculation. Starting main diffusion equation calculations over time and frequency..."
    )

    # %%
    ###############################################################################
    # CALCULATION OF EQUIVALENT ABSORPTION AREA
    ###############################################################################
    def equiv_absorp():
        V = sum(cell_volume)  # volume of the room
        S = total_boundArea  # total surface area of the room

        sum_alpha_average = 0
        Eq_A = 0
        # Absorption parameters for room
        for entity in surface_areas:
            # print(entity)
            sum_alpha_average += np.multiply(
                absorption_coefficient[entity], surface_areas[entity]
            )
            Eq_A += np.multiply(absorption_coefficient[entity], surface_areas[entity])
        # alpha_average = sum_alpha_average/S #average absorption

        return V, S, Eq_A

    # FUNCTION CALLED HERE
    V, S, Eq_A = equiv_absorp()

    # %%
    ###############################################################################
    # CALCULATION OF SABINE RT, TOTAL RECORDING TIME AND SOURCE ON TIME
    ###############################################################################

    def rec_time():
        RT_Sabine_band = []
        for iBand in range(nBands):
            # freq = center_freq[iBand]
            RT_Sabine = 0.16 * V / Eq_A[iBand]
            RT_Sabine_band.append(RT_Sabine)

        sourceon_time = round(
            max(RT_Sabine_band), 1
        )  # time that the source is ON before interrupting [s]
        if result_container:
            if use_default:
                recording_time = sourceon_time + (sourceon_time / 60 * 35)
            else:
                if simulation_settings["sim_len_type"] == "ir_length":
                    recording_time = sourceon_time + simulation_settings["de_ir_length"]
                else:
                    recording_time = sourceon_time + (
                        sourceon_time / 60 * simulation_settings["edt"]
                    )
        else:
            recording_time = (
                2 * sourceon_time
            )  # total time recorded for the calculation [s]

        # Time resolution
        t = np.arange(0, recording_time, dt)  # mesh point in time
        recording_steps = ceil(
            recording_time / dt
        )  # number of time steps to consider in the calculation

        return RT_Sabine_band, sourceon_time, recording_time, t, recording_steps

    # FUNCTION CALLED HERE
    RT_Sabine_band, sourceon_time, recording_time, t, recording_steps = rec_time()

    print("Source on time: " + str(sourceon_time))
    print("Total recording time: " + str(recording_time))

    # %%
    ###############################################################################
    # CALCULATION OF DIFFUSION PARAMETERS
    ###############################################################################
    # Diffusion parameters
    def diff_coeff():
        mean_free_path = (4 * V) / S  # mean free path for 3D
        # mean_free_time = mean_free_path/c0 #mean free time for 3D
        # mean_free_time_step = int(mean_free_time/dt)
        Dx = (
            mean_free_path * c0
        ) / 3  # diffusion coefficient for proportionate rooms x direction
        Dy = (
            mean_free_path * c0
        ) / 3  # diffusion coefficient for proportionate rooms y direction
        Dz = (
            mean_free_path * c0
        ) / 3  # diffusion coefficient for proportionate rooms z direction
        return Dx, Dy, Dz

    # FUNCTION CALLED HERE
    Dx, Dy, Dz = diff_coeff()

    # %%
    ###############################################################################
    # CALCULATION OF SOURCE & RECEIVER DISTANCE
    ###############################################################################

    def dist_source_receiver():
        # distance between source and receiver
        dist_sr = math.sqrt(
            (abs(coord_rec[0] - coord_source[0])) ** 2
            + (abs(coord_rec[1] - coord_source[1])) ** 2
            + (abs(coord_rec[2] - coord_source[2])) ** 2
        )  # distance between source and receiver
        return dist_sr

    # FUNCTION CALLED HERE
    dist_sr = dist_source_receiver()

    # %%
    ###############################################################################
    # SOURCE INTERPOLATION
    ###############################################################################
    def source_interp():
        # ORIGINAL
        # Position of source is the centre of a cell so the minimum distance with the centre of a cell has been calculated to understand which cell is the closest
        # dist_source_cc_list = []
        # for i in range(len(cell_center)):
        #     dist_source_cc = math.sqrt(np.sum((cell_center[i] - coord_source)**2))
        #     dist_source_cc_list.append(dist_source_cc)
        # source_idx = np.argmin(dist_source_cc_list)

        # SOURCE INTERPOLATION CALCULATED WITHIN N CENTRE CELL SELECTED WITHIN THE LENGTH OF MESH
        # Position of source is the centre of a cell so the minimum distance with the centre of a cell has been calculated to understand which cell is the closest
        # dist_source_cc_list = []
        # for i in range(len(cell_center)):
        #     dist_source_cc = math.sqrt(np.sum((cell_center[i] - coord_source)**2))
        #     dist_source_cc_list.append(dist_source_cc)
        # source_idx = np.argmin(dist_source_cc_list)
        # import math
        # #cl_tet_s stands for cl=closest, tet=tetrahedron, s=to the source
        # cl_tet_s = {} #initialise dictionary fro closest tetrahedrons to the source
        # for i in range(len(dist_source_cc_list)):
        #     if dist_source_cc_list[i] < length_of_mesh: #the number of closest tetrahedron to the source depends on the mesh and I would say that 0.5 should be equal to the length of mesh
        #         cl_tet_s[i] = dist_source_cc_list[i]
        # total_weights_s = {} #initialise weights for each tetrahedron around the actual source position
        # sum_weights_s = 0
        # Vs = 0
        # for i, dist in cl_tet_s.items(): #for each key and value in the dictionary (so for each closest tetrahedron to the source)
        #     weights = np.divide(1.0 , dist)  #calculate the inverse distance weights, so closer to the point means higher weight
        #     #print(weights)
        #     sum_weights_s += weights
        #     #weights /= np.sum(weights)  # Normalize weights to sum to 1
        #     total_weights_s[i] = weights #put the wweigths (values) to the correspondent closest tetrahedron (keys)
        # #    Vs += cell_volume[i] #volume of the source calculated summing the volumes of all the tetrahedrons involved
        # #total_weights_s_values = total_weights_s.values()
        # for i,weight in total_weights_s.items():
        #     total_weights_s[i] = weight/sum_weights_s if sum_weights_s != 0 else 0
        # cl_tet_s_keys = cl_tet_s.keys() #take only the keys of the cl_tet_s dictionary (so basically the indexes of the tetrahedrons)

        ################SOURCE INTERPOLATION WITH VERTICES OF TETRAHEDRON IN WHICH SOURCE IS IN
        # Position of source is the centre of a cell so the minimum distance with the centre of a cell has been calculated to understand which cell is the closest
        # dist_source_cc_list = []
        # for i in range(len(cell_center)):
        #     dist_source_cc = math.sqrt(np.sum((cell_center[i] - coord_source)**2))
        #     dist_source_cc_list.append(dist_source_cc)
        # source_idx = np.argmin(dist_source_cc_list)

        # node_source_in = voluNode[source_idx]
        # vertices_source_in = np.array([]).reshape(0, 3)
        # for vx in node_source_in:
        #     vertices = nodecoords[vx]
        #     vertices_source_in = np.vstack((vertices_source_in, vertices))

        # dist_source_vx_list = []
        # for i in range(len(vertices_source_in)):
        #     dist_source_vx = math.sqrt(np.sum((vertices_source_in[i] - coord_source)**2))
        #     dist_source_vx_list.append(dist_source_vx)

        # total_weights_s = {} #initialise weights for the tetrahedron around the actual source position
        # sum_weights_s = 0
        # Vs = 0
        # for i in range(len(dist_source_vx_list)): #for each key and value in the dictionary (so for each closest tetrahedron to the source)
        #     weights = np.divide(1.0 , dist_source_vx_list[i])  #calculate the inverse distance weights, so closer to the point means higher weight
        #     #print(weights)
        #     sum_weights_s += weights
        #     #weights /= np.sum(weights)  # Normalize weights to sum to 1
        #     total_weights_s[i] = weights #put the wweigths (values) to the correspondent closest tetrahedron (keys)
        #     #Vs += cell_volume[i] #volume of the source calculated summing the volumes of all the tetrahedrons involved

        # #total_weights_s_values = total_weights_s.values()
        # for i,weight in total_weights_s.items():
        #     total_weights_s[i] = weight/sum_weights_s if sum_weights_s != 0 else 0

        # Vs = cell_volume[source_idx] #volume of the source = to volume of cells where the volume is

        # SOURCE INTERPOLATION CALCULATED WITHIN 4 CENTRE CELL SELECTED (TETRAHEDRON)
        # Position of source is the centre of a cell so the minimum distance with the centre of a cell has been calculated to understand which cell is the closest
        dist_source_cc_list = (
            []
        )  # initialise the list for all the distances between each cell centre and the source
        for i in range(len(cell_center)):  # for each tetra
            dist_source_cc = math.sqrt(
                np.sum((cell_center[i] - coord_source) ** 2)
            )  # calculate the distance between its centre cell and the source coordinate
            dist_source_cc_list.append(dist_source_cc)  # append the distance in a list
        # source_idx = np.argmin(dist_source_cc_list) #take the minimum distance index; this is where the source will be positioned

        dist_source_cc_list_sorted = sorted(
            dist_source_cc_list
        )  # sorted from the minimum to the maximum distance
        selected_source_cc_list = dist_source_cc_list_sorted[
            :4
        ]  # take only the first four element of the sorted list (take the first 4 cell centres closest to the source)

        dist_source_cc_list_sorted_indices = np.argsort(dist_source_cc_list)[
            :4
        ]  # takes the indeces of the minimum distances
        # selected_source_cc_list_indices = dist_source_cc_list_sorted_indices[:4] #does exactly the same as the previous line

        # cl_tet_s stands for cl=closest, tet=tetrahedron, s=to the source
        cl_tet_s = {}  # initialise dictionary for closest tetrahedrons to the source
        for i in range(
            len(dist_source_cc_list_sorted_indices)
        ):  # for each of the 4 tetra closest to the source
            idx = selected_source_cc_list[i]  # take its index
            cl_tet_s[dist_source_cc_list_sorted_indices[i]] = (
                idx  # put it in a dictionary
            )

        total_weights_s = (
            {}
        )  # initialise weights for each tetrahedron around the actual source position
        sum_weights_s = 0
        # Vs = 0
        for (
            i,
            dist,
        ) in (
            cl_tet_s.items()
        ):  # for each key and value in the dictionary (so for each closest tetrahedron to the source)
            weights = np.divide(
                1.0, dist
            )  # calculate the inverse distance weights, so closer to the point means higher weight
            # print(weights)
            sum_weights_s += weights
            # weights /= np.sum(weights)  # Normalize weights to sum to 1
            total_weights_s[i] = (
                weights  # put the wweigths (values) to the correspondent closest tetrahedron (keys)
            )
            # Vs += cell_volume[i] #volume of the source calculated summing the volumes of all the tetrahedrons involved

        # total_weights_s_values = total_weights_s.values()
        for i, weight in total_weights_s.items():
            total_weights_s[i] = weight / sum_weights_s if sum_weights_s != 0 else 0

        cl_tet_s_keys = (
            cl_tet_s.keys()
        )  # take only the keys of the cl_tet_s dictionary (so basically the indexes of the tetrahedrons)

        return cl_tet_s_keys, total_weights_s

    # FUNCTION CALLED HERE
    cl_tet_s_keys, total_weights_s = source_interp()

    # %%
    ###############################################################################
    # CALCULATION OF SOURCE VOLUME
    ###############################################################################

    def source_volume():
        # To make sure that the source is in the correct tetrahedron position
        node_ids = velemNodes.T
        ori = nodecoords[node_ids[0, :] - 1, :]
        # ori = nodecoords[node_indices[node_ids[:,0],:]]
        v_tet_s1 = nodecoords[node_ids[1, :] - 1, :] - ori
        v_tet_s2 = nodecoords[node_ids[2, :] - 1, :] - ori
        v_tet_s3 = nodecoords[node_ids[3, :] - 1, :] - ori
        n_tet = len(node_ids.T)
        v1s = v_tet_s1.T.reshape((3, 1, n_tet))
        v2s = v_tet_s2.T.reshape((3, 1, n_tet))
        v3s = v_tet_s3.T.reshape((3, 1, n_tet))
        mat = np.concatenate((v1s, v2s, v3s), axis=1)
        inv_mat = np.linalg.inv(mat.T).T
        # if coord_source.size==3:  # to make rec has a dimension of (N_rec,3)
        #    rec=rec.reshape((1,3))
        coord_source_array = np.array(coord_source)
        if coord_source_array.size == 3:  # to make rec has a dimension of (N_rec,3)
            coord_source_array = coord_source_array.reshape((1, 3))
        N_sou = coord_source_array.shape[0]
        orir = np.repeat(ori[:, :, np.newaxis], N_sou, axis=2)
        newp = np.einsum("imk,kmj->kij", inv_mat, coord_source_array.T - orir)
        val = (
            np.all(newp >= 0, axis=1)
            & np.all(newp <= 1, axis=1)
            & (np.sum(newp, axis=1) <= 1)
        )
        id_tet, id_p = np.nonzero(val)
        res = -np.ones(N_sou, dtype=id_tet.dtype)  # Sentinel value
        res[id_p] = id_tet

        ###############################################################################
        # #VOLUME CALCULATED WITHIN 4 CENTRE CELL SELECTED (TETRAHEDRON)
        # Calculate volume of the source with the 4 cell centres as the vertices of the tetrahedron
        # vertices_source = np.array([]).reshape(0, 3)  # Initialize as an empty 2D array with 3 columns
        # for tet in cl_tet_s_keys:
        #     vertices = cell_center[tet]
        #     vertices_source = np.vstack((vertices_source, vertices))

        # # Calculate edge vectors
        # AB = vertices_source[0] - vertices_source[3]
        # CD = vertices_source[1] - vertices_source[3]
        # EF = vertices_source[2] - vertices_source[3]

        # Calculate volume using scalar triple product
        # Vs = np.abs(np.dot(AB, np.cross(CD, EF))) / 6.0

        # VOLUME CALCULATED WITHIN N CENTRE CELL SELECTED
        # vertices_source = np.array([]).reshape(0, 3)  # Initialize as an empty 2D array with 3 columns
        # for tet in cl_tet_s_keys:
        #     vertices = cell_center[tet]
        #     vertices_source = np.vstack((vertices_source, vertices))
        # from scipy.spatial import ConvexHull

        # def calculate_volume(vertices_source):
        #     # Create a ConvexHull object
        #     hull = ConvexHull(vertices_source)

        #     # Calculate the volume of the convex hull
        #     volume = hull.volume

        #     return volume
        # Vs = calculate_volume(vertices_source)

        # VOLUME ORIGINAL
        Vs = cell_volume[
            res[0]
        ]  # volume of the source = to volume of cell where the source is
        # Vs = 1

        return Vs

    # FUNCTION CALLED HERE
    Vs = source_volume()

    # %%
    ###############################################################################
    # INITIAL CONDITIONS
    ###############################################################################
    # Initial condition - Source Info (interrupted method)
    def initial_cond():
        w1 = Ws / Vs  # w1 = round(Ws/Vs,4) #power density of the source [Watts/(m^3))]
        sourceon_steps = ceil(
            sourceon_time / dt
        )  # time steps at which the source is calculated/considered in the calculation
        s1 = np.multiply(
            w1, np.ones(sourceon_steps)
        )  # energy density of source number 1 at each time step position
        source1 = np.append(
            s1, np.zeros(recording_steps - sourceon_steps)
        )  # This would be equal to s1 if and only if recoding_steps = sourceon_steps
        return source1, sourceon_steps

    # FUNCTION CALLED HERE
    source1, sourceon_steps = initial_cond()

    # %%
    ###############################################################################
    # DEFINITION OF SOURCE MATRIX
    ###############################################################################
    def source_matrix():
        # ORIGINAL SOURCE MATRIX
        # s = np.zeros((len(voluEl))) #matrix of zeros for source
        # s[source_idx] = source1[0]

        # INTERPOLATION WITH CELL CENTRES - SOURCE MATRIX
        s = np.zeros((len(voluEl)))  # matrix of zeros for source
        for tet_s in cl_tet_s_keys:
            s[tet_s] = source1[0] * total_weights_s[tet_s]

        # INTERPOLATION WITH VERTICES - SOURCE MATRIX
        # s = np.zeros((len(voluEl))) #matrix of zeros for source
        # for i in total_weights_s:
        #    s[source_idx] += source1[0] *total_weights_s[i]

        return s

    # FUNCTION CALLED HERE
    s = source_matrix()

    # %%
    ###############################################################################
    # RECEIVER INTERPOLATION
    ###############################################################################
    def receiver_interp():
        # ORIGINAL
        # dist_rec_cc_list = []
        # for i in range(len(cell_center)):
        #     dist_rec_cc = math.sqrt(np.sum((cell_center[i] - coord_rec)**2))
        #     dist_rec_cc_list.append(dist_rec_cc)
        # rec_idx = np.argmin(dist_rec_cc_list)

        # INTERPOLATION WITH CLOSEST CENTRES WITHIN THE LENGTH OF MESH
        # Position of receiver is the centre of a cell so the minimum distance with the centre of a cell has been calculated to understand which cell is the closest
        # dist_rec_cc_list = []
        # for i in range(len(cell_center)):
        #     dist_rec_cc = math.sqrt(np.sum((cell_center[i] - coord_rec)**2))
        #     dist_rec_cc_list.append(dist_rec_cc)
        # rec_idx = np.argmin(dist_rec_cc_list)

        # #cl_tet_r stands for cl=closest, tet=tetrahedron, r=to the receiver
        # cl_tet_r = {} #initialise dictionary fro closest tetrahedrons to the source
        # for i in range(len(dist_rec_cc_list)):
        #     if dist_rec_cc_list[i] < length_of_mesh: #the number of closest tetrahedron to the source depends on the mesh and I would say that 0.5 should be equal to the length of mesh
        #         cl_tet_r[i] = dist_rec_cc_list[i]

        # total_weights_r = {} #initialise weights for each tetrahedron around the actual source position
        # sum_weights_r = 0
        # for i, dist in cl_tet_r.items(): #for each key and value in the dictionary (so for each closest tetrahedron to the source)
        #     weights = np.divide(1.0 , dist)  #calculate the inverse distance weights, so closer to the point means higher weight
        #     #print(weights)
        #     sum_weights_r += weights
        #     #weights /= np.sum(weights)  # Normalize weights to sum to 1
        #     total_weights_r[i] = weights #put the wweigths (values) to the correspondent closest tetrahedron (keys)

        # #total_weights_s_values = total_weights_s.values()
        # for i,weight in total_weights_r.items():
        #     total_weights_r[i] = weight/sum_weights_r if sum_weights_r != 0 else 0

        # cl_tet_r_keys = cl_tet_r.keys() #take only the keys of the cl_tet_s dictionary (so basically the indexes of the tetrahedrons)

        # RECEIVER INTERPOLATION CALCULATED WITHIN 4 CENTRE CELL SELECTED (TETRAHEDRON)
        # Position of receiver is the centre of a cell so the minimum distance with the centre of a cell has been calculated to understand which cell is the closest
        dist_rec_cc_list = (
            []
        )  # initialise the list for all the distances between each cell centre and the source
        for i in range(len(cell_center)):  # for each tetra
            dist_rec_cc = math.sqrt(
                np.sum((cell_center[i] - coord_rec) ** 2)
            )  # calculate the distance between its centre cell and the source coordinate
            dist_rec_cc_list.append(dist_rec_cc)  # append the distance in a list
        # rec_idx = np.argmin(dist_rec_cc_list) #take the minimum distance index; this is where the source will be positioned

        dist_rec_cc_list_sorted = sorted(
            dist_rec_cc_list
        )  # sorted from the minimum to the maximum distance
        selected_rec_cc_list = dist_rec_cc_list_sorted[
            :4
        ]  # take only the first four element of the sorted list (take the first 4 cell centres closest to the source)

        dist_rec_cc_list_sorted_indices = np.argsort(dist_rec_cc_list)[
            :4
        ]  # takes the indeces of the minimum distances
        # selected_rec_cc_list_indices = dist_rec_cc_list_sorted_indices[:4] #does exactly the same as the previous line

        # cl_tet_s stands for cl=closest, tet=tetrahedron, s=to the source
        cl_tet_r = {}  # initialise dictionary for closest tetrahedrons to the source
        for i in range(
            len(dist_rec_cc_list_sorted_indices)
        ):  # for each of the 4 tetra closest to the source
            idx = selected_rec_cc_list[i]  # take its index
            cl_tet_r[dist_rec_cc_list_sorted_indices[i]] = idx  # put it in a dictionary

        total_weights_r = (
            {}
        )  # initialise weights for each tetrahedron around the actual source position
        sum_weights_r = 0
        # Vs = 0
        for (
            i,
            dist,
        ) in (
            cl_tet_r.items()
        ):  # for each key and value in the dictionary (so for each closest tetrahedron to the source)
            weights = np.divide(
                1.0, dist
            )  # calculate the inverse distance weights, so closer to the point means higher weight
            # print(weights)
            sum_weights_r += weights
            # weights /= np.sum(weights)  # Normalize weights to sum to 1
            total_weights_r[i] = (
                weights  # put the wweigths (values) to the correspondent closest tetrahedron (keys)
            )
            # Vs += cell_volume[i] #volume of the source calculated summing the volumes of all the tetrahedrons involved

        # total_weights_s_values = total_weights_s.values()
        for i, weight in total_weights_r.items():
            total_weights_r[i] = weight / sum_weights_r if sum_weights_r != 0 else 0

        cl_tet_r_keys = (
            cl_tet_r.keys()
        )  # take only the keys of the cl_tet_s dictionary (so basically the indexes of the tetrahedrons)

        return cl_tet_r_keys, total_weights_r

    # FUNCTION CALLED HERE
    cl_tet_r_keys, total_weights_r = receiver_interp()

    # %%
    ###############################################################################
    # CALCULATION OF LENGTH, WIDTH AND HEIGHT OF ROOM
    ###############################################################################

    def room_dimensions():
        # Extract x-coordinates of all nodes LENGTH
        x_coordinates = nodecoords[:, 0]
        # Find the minimum and maximum x-coordinates to determine the length of the room
        min_x = np.min(x_coordinates)
        max_x = np.max(x_coordinates)
        # Calculate the length of the room
        room_length = max_x - min_x

        # Extract y-coordinates of all nodes WIDTH
        y_coordinates = nodecoords[:, 1]
        # Find the minimum and maximum x-coordinates to determine the width of the room
        min_y = np.min(y_coordinates)
        max_y = np.max(y_coordinates)
        # Calculate the width of the room
        room_width = max_y - min_y

        # Extract y-coordinates of all nodes HEIGHT
        z_coordinates = nodecoords[:, 2]
        # Find the minimum and maximum x-coordinates to determine the width of the room
        min_z = np.min(z_coordinates)
        max_z = np.max(z_coordinates)
        # Calculate the height of the room
        room_height = max_z - min_z

        return room_length, room_width, room_height

    # FUNCTION CALLED HERE
    room_length, room_width, room_height = room_dimensions()

    # %%
    ###############################################################################
    # CALCULATION OF LINE RECEIVERS
    ###############################################################################
    def line_receivers():
        # Arange linespace lines
        dx = 0.5
        x_axis = np.arange(
            0, room_length + dx, dx
        )  # lispace on x_axis with distance dx
        y_axis = np.arange(0, room_width + dx, dx)

        # RECEIVERS IN A X LINE
        line_rec_x_idx_list = []
        dist_x = np.array([])
        for x_chang in x_axis:
            line_rec = [x_chang, coord_rec[1], coord_rec[2]]
            # Position of line_receiver is the centre of a cell
            dist_line_rec_x = math.sqrt(
                (abs(line_rec[0] - coord_source[0])) ** 2
                + (abs(line_rec[1] - coord_source[1])) ** 2
                + (abs(line_rec[2] - coord_source[2])) ** 2
            )  # distance between source and line_receiver
            dist_x = np.append(dist_x, dist_line_rec_x)  # Append to the NumPy array
            dist_line_rec_x_cc_list = []
            for i in range(len(cell_center)):
                dist_line_rec_x_cc = math.sqrt(np.sum((cell_center[i] - line_rec) ** 2))
                dist_line_rec_x_cc_list.append(dist_line_rec_x_cc)
            line_rec_x_idx = np.argmin(dist_line_rec_x_cc_list)
            line_rec_x_idx_list.append(line_rec_x_idx)

            # RECEIVERS IN A Y LINE
        line_rec_y_idx_list = []
        dist_y = np.array([])
        for y_chang in y_axis:
            line_rec = [coord_rec[0], y_chang, coord_rec[2]]
            # Position of line_receiver is the centre of a cell
            dist_line_rec_y = math.sqrt(
                (abs(line_rec[0] - coord_source[0])) ** 2
                + (abs(line_rec[1] - coord_source[1])) ** 2
                + (abs(line_rec[2] - coord_source[2])) ** 2
            )  # distance between source and line_receiver
            dist_y = np.append(dist_y, dist_line_rec_y)  # Append to the NumPy array
            dist_line_rec_y_cc_list = []
            for i in range(len(cell_center)):
                dist_line_rec_y_cc = math.sqrt(np.sum((cell_center[i] - line_rec) ** 2))
                dist_line_rec_y_cc_list.append(dist_line_rec_y_cc)
            line_rec_y_idx = np.argmin(dist_line_rec_y_cc_list)
            line_rec_y_idx_list.append(line_rec_y_idx)

            # def interpolate_receiver_position(interior_tet_tet, cell_centers, receiver_position):
        #    distances = np.sqrt(np.sum((cell_centers - receiver_position)**2, axis=1))
        #    weights = interior_tet_tet[rec_idx, :] / distances  # Use the row corresponding to the receiver index
        #    interpolated_position = np.dot(weights, cell_centers) / np.sum(weights)
        #    return interpolated_position

        # interpolated_receiver_position = interpolate_receiver_position(interior_tet_tet, cell_center, coord_rec)

        # dist_sr_interpolated = np.linalg.norm(interpolated_receiver_position - coord_source)

        return x_axis, y_axis, line_rec_x_idx_list, dist_x, line_rec_y_idx_list, dist_y

    # FUNCTION CALLED HERE
    x_axis, y_axis, line_rec_x_idx_list, dist_x, line_rec_y_idx_list, dist_y = (
        line_receivers()
    )

    # %%
    ###############################################################################
    # CALCULATION OF BETA_ZERO
    ###############################################################################

    def beta_zero():
        beta_zero_freq = []
        for iBand in range(len(boundary_areas)):
            # print(iBand)
            # freq = center_freq[iBand]
            # print(boundary_areas[iBand])
            beta_zero_element = np.divide(
                dt * ((Dx * interior_tet_sum) + boundary_areas[iBand]), cell_volume
            )  # my interpretation of the beta_zero
            beta_zero_freq.append(beta_zero_element)

        return beta_zero_freq

    # FUNCTION CALLED HERE
    beta_zero_freq = beta_zero()

    # %%
    ###############################################################################
    # MAIN CALCULATION - COMPUTING ENERGY DENSITY
    ###############################################################################

    def computing_energy_density():
        w_new_band = []
        w_rec_band = []
        w_rec_off_band = []
        w_rec_off_deriv_band = []
        p_rec_off_deriv_band = []

        prevPercentDone = 0

        for iBand in range(nBands):

            if check_should_cancel(json_file_path):
                break

            w_new = np.zeros(len(voluEl))  # unknown w at new time level (n+1)
            # w_old = np.zeros(len(voluEl))
            w = w_new  # w at n level
            w_old = w  # w_old at n-1 level
            # w[source_idx] = w1 #w (m time step) at source position -> impulse source

            w_rec = np.zeros(recording_steps)  # energy density at the receiver

            # Computing w;
            for steps in range(0, recording_steps):
                # Compute w at inner mesh points
                # time_steps = steps*dt #total time for the calculation

                # Computing w_new (w at n+1 time step)

                w_new = (
                    np.divide(
                        (np.multiply(w_old, (1 - beta_zero_freq[iBand]))),
                        (1 + beta_zero_freq[iBand]),
                    )
                    - np.divide((2 * dt * c0 * m_atm * w), (1 + beta_zero_freq[iBand]))
                    + np.divide(
                        np.divide((2 * dt * Dx * (interior_tet @ w)), cell_volume),
                        (1 + beta_zero_freq[iBand]),
                    )
                    + np.divide((2 * dt * s), (1 + beta_zero_freq[iBand]))
                )  # The absorption term is part of beta_zero

                # Update w before next step
                w_old = w  # The w at n step becomes the w at n-1 step
                w = w_new  # The w at n+1 step becomes the w at n step

                # w_new_total = 0
                # w_rec is the energy density at the specific receiver
                # ORIGINAL
                # w_rec[steps] = w_new[rec_idx]

                # INTERPOLATION WITH N CELL CENTRES OR 4 CELL CENTRES
                for tet_r in cl_tet_r_keys:
                    w_rec[steps] += w_new[tet_r] * total_weights_r[tet_r]

                    # if steps == sourceon_steps:
                #     #print("Steps for source:",steps)
                #     w_t0 = w_new

                # if steps == round(1*mean_free_time_step + sourceon_steps + (dist_sr/c0)):
                #     w_1l = w_new

                # if steps == round(2*mean_free_time_step + sourceon_steps + (dist_sr/c0)):
                #     w_2l = w_new

                # if steps == round(3*mean_free_time_step + sourceon_steps + (dist_sr/c0)):
                #     w_3l = w_new

                # if steps == round(5*mean_free_time_step + sourceon_steps + (dist_sr/c0)):
                #     w_5l = w_new

                if tcalc == "decay":
                    # INTERPOLATION WITH N CELL CENTRES OR 4 CELL CENTRES
                    for tet_s in cl_tet_s_keys:
                        s[tet_s] = source1[steps] * total_weights_s[tet_s]
                    # ORIGINAL
                    # s[source_idx] = source1[steps]
                    # INTERPOLATION WITH VERTICES
                    # for i in total_weights_s:
                    #    s[source_idx] = source1[steps] *total_weights_s[i]

                if tcalc == "stationarysource":
                    # ORIGINAL
                    # s[source_idx] = source1[0]

                    # INTERPOLATION SOURCE
                    for tet_s in cl_tet_s_keys:
                        s[tet_s] = source1[0] * total_weights_s[tet_s]

                idx_w_rec = np.argmin(
                    np.abs(t - sourceon_time)
                )  # index at which the t array is equal to the sourceon_time; I want the RT to calculate from when the source stops.
                w_rec_off = w_rec[idx_w_rec:]
                p_rec_off = (w_rec[idx_w_rec:]) * rho * c0**2
                t_off = t[idx_w_rec:]

                # Envelope of Impulse response from the energy density
                w_rec_off_deriv = w_rec_off  # initialising an array of derivative equal to the w_rec_off -> this will be the impulse response after modifying it
                w_rec_off_deriv = np.delete(
                    w_rec_off_deriv, 0
                )  # delete the first element of the array -> this means shifting the array one step before and therefore do a derivation
                w_rec_off_deriv = np.append(
                    w_rec_off_deriv, 0
                )  # add a zero in the last element of the array -> for derivation and to have the same length as previously
                # impulse = ((w_rec_off - w_rec_off_deriv))/dt#/(rho*c0**2) #This is the difference between the the energy density and the impulse response

                # Envelope of Impulse response from the pressure
                p_rec_off_deriv = p_rec_off  # initialising an array of derivative equal to the w_rec_off -> this will be the impulse response after modifying it
                p_rec_off_deriv = np.delete(
                    p_rec_off_deriv, 0
                )  # delete the first element of the array -> this means shifting the array one step before and therefore do a derivation
                p_rec_off_deriv = np.append(
                    p_rec_off_deriv, 0
                )  # add a zero in the last element of the array -> for derivation and to have the same length as previously

                # print(time_steps)

                percentDone = round(
                    100 * (iBand / nBands + steps / recording_steps * 1 / nBands)
                )
                if percentDone > prevPercentDone:
                    # Checking whether the user has cancelled the simulation (only one time per percentage increase)
                    if check_should_cancel(json_file_path):
                        print("breaking out of inner loop")
                        break

                    print(str(percentDone) + "% of main calculation completed")
                    if result_container:
                        result_container["results"][0]["percentage"] = percentDone
                        with open(json_file_path, "w") as percentage_update:
                            percentage_update.write(
                                json.dumps(result_container, indent=4)
                            )

                prevPercentDone = percentDone

            w_new_band.append(w_new)
            w_rec_band.append(w_rec)
            w_rec_off_band.append(w_rec_off)
            w_rec_off_deriv_band.append(w_rec_off_deriv)
            p_rec_off_deriv_band.append(p_rec_off_deriv)

            if check_should_cancel(json_file_path):
                print("breaking out of outer loop")
                break

        if check_should_cancel(json_file_path):
            print("returning empty data")
            return [], [], [], [], [], 0, []

        return (
            w_new_band,
            w_rec_band,
            w_rec_off_band,
            w_rec_off_deriv_band,
            p_rec_off_deriv_band,
            idx_w_rec,
            t_off,
        )

    # FUNCTION CALLED HERE
    (
        w_new_band,
        w_rec_band,
        w_rec_off_band,
        w_rec_off_deriv_band,
        p_rec_off_deriv_band,
        idx_w_rec,
        t_off,
    ) = computing_energy_density()

    if check_should_cancel(json_file_path):
        print("returning to simulation_service")
        return

    print("100% of main calculation completed")
    if result_container:
        result_container["results"][0]["percentage"] = 100
        with open(json_file_path, "w") as percentage_update:
            percentage_update.write(json.dumps(result_container, indent=4))

    # %%
    ###############################################################################
    # POST-PROCESS
    ###############################################################################

    def freq_parameters():
        w_rec_x_band = []
        w_rec_y_band = []
        spl_stat_x_band = []
        spl_stat_y_band = []
        spl_r_band = []
        spl_r_off_band = []
        spl_r_norm_band = []
        t20_band = []
        t30_band = []
        edt_band = []
        c80_band = []
        d50_band = []
        ts_band = []
        sch_db_band = []
        spl_r_t0_band = []

        for iBand in range(nBands):

            w_rec_x_end = np.array([])
            for xr in line_rec_x_idx_list:
                w_rec_x = w_new_band[iBand][xr]
                w_rec_x_end = np.append(w_rec_x_end, w_rec_x)

            w_rec_y_end = np.array([])
            for yr in line_rec_y_idx_list:
                w_rec_y = w_new_band[iBand][yr]
                w_rec_y_end = np.append(w_rec_y_end, w_rec_y)

            spl_stat_x = 10 * np.log10(
                rho
                * c0
                * (((Ws) / (4 * math.pi * (dist_x**2))) + ((abs(w_rec_x_end) * c0)))
                / (pRef**2)
            )
            spl_stat_y = 10 * np.log10(
                rho
                * c0
                * (((Ws) / (4 * math.pi * (dist_y**2))) + ((abs(w_rec_y_end) * c0)))
                / (pRef**2)
            )  # It should be the spl stationary

            # press_r = ((abs(w_rec_band[iBand]))*rho*(c0**2)) #pressure at the receiver
            spl_r = 10 * np.log10(
                ((abs(w_rec_band[iBand])) * rho * (c0**2)) / (pRef**2)
            )  # ,where=press_r>0, sound pressure level at the receiver
            spl_r_off = 10 * np.log10(
                ((abs(w_rec_off_band[iBand])) * rho * (c0**2)) / (pRef**2)
            )

            spl_r_t0 = spl_r_off[0]

            spl_r_norm = 10 * np.log10(
                (((abs(w_rec_band[iBand])) * rho * (c0**2)) / (pRef**2))
                / np.max(((abs(w_rec_band[iBand])) * rho * (c0**2)) / (pRef**2))
            )  # normalised to maximum to 0dB
            # spl_r_tot = 10*np.log10(rho*c0*((Ws/(4*math.pi*dist_sr**2))*np.exp(-m_atm*dist_sr) + ((abs(w_rec_band[iBand]))*c0))/(pRef**2)) #spl total (including direct field) at the receiver position????? but it will need to be calculated for a stationary source 100dB

            # Find the energy decay part of the overal calculation
            # idx_w_rec = np.argmin(np.abs(t - sourceon_time)) #index at which the t array is equal to the sourceon_time; I want the RT to calculate from when the source stops.
            # idx_w_rec = np.where(t == sourceon_time)[0][0] #index at which the t array is equal to the sourceon_time; I want the RT to calculate from when the source stops.
            # w_rec_off = w_rec_band[iBand][idx_w_rec:] #cutting the energy density array at the receiver from the idx_w_rec to the end

            # Schroeder integration
            # energy_r_rev = (w_rec_off)[::-1] #reverting the array
            # The energy density is related to the pressure with the following relation: w = p^2
            # energy_r_rev_cum = np.cumsum(energy_r_rev) #cumulative summation of all the item in the array
            schroeder = w_rec_off_band[
                iBand
            ]  # energy_r_rev_cum[::-1] #reverting the array again -> creating the schroder decay
            sch_db = 10.0 * np.log10(
                schroeder / max(schroeder)
            )  # level of the array: schroeder decay

            if tcalc == "decay":
                t20 = t60_decay(
                    t, sch_db, idx_w_rec, rt="t20"
                )  # called function for calculation of t20 [s]
                t30 = t60_decay(
                    t, sch_db, idx_w_rec, rt="t30"
                )  # called function for calculation of t60 [s]
                edt = t60_decay(
                    t, sch_db, idx_w_rec, rt="edt"
                )  # called function for calculation of edt [s]
                # Eq_A = 0.16*V/t60 #equivalent absorption area defined from the RT
                c80 = clarity(
                    t30, V, Eq_A[iBand], S, c0, dist_sr
                )  # called function for calculation of c80 [dB]
                d50 = definition(
                    t30, V, Eq_A[iBand], S, c0, dist_sr
                )  # called function for calculation of d50 [%]
                ts = centretime(
                    t30, Eq_A[iBand], S
                )  # called function for calculation of ts [ms]

                t20_band.append(t20)
                t30_band.append(t30)
                edt_band.append(edt)
                c80_band.append(c80)
                d50_band.append(d50)
                ts_band.append(ts)

            w_rec_x_band.append(w_rec_x_end)
            w_rec_y_band.append(w_rec_y_end)
            spl_stat_x_band.append(spl_stat_x)
            spl_stat_y_band.append(spl_stat_y)
            spl_r_band.append(spl_r)
            spl_r_off_band.append(spl_r_off)
            spl_r_norm_band.append(spl_r_norm)
            sch_db_band.append(sch_db)
            spl_r_t0_band.append(spl_r_t0)

        spl_r_off_band = np.array(spl_r_off_band)
        spl_r_t0_band = np.array(spl_r_t0_band)
        t20_band = np.array(t20_band)
        t30_band = np.array(t30_band)
        edt_band = np.array(edt_band)
        c80_band = np.array(c80_band)
        d50_band = np.array(d50_band)
        ts_band = np.array(ts_band)

        return (
            w_rec_x_band,
            w_rec_y_band,
            spl_stat_x_band,
            spl_stat_y_band,
            spl_r_band,
            spl_r_off_band,
            spl_r_norm_band,
            sch_db_band,
            t20_band,
            t30_band,
            edt_band,
            c80_band,
            d50_band,
            ts_band,
            spl_r_t0_band,
        )

    if not check_should_cancel(json_file_path):
        (
            w_rec_x_band,
            w_rec_y_band,
            spl_stat_x_band,
            spl_stat_y_band,
            spl_r_band,
            spl_r_off_band,
            spl_r_norm_band,
            sch_db_band,
            t20_band,
            t30_band,
            edt_band,
            c80_band,
            d50_band,
            ts_band,
            spl_r_t0_band,
        ) = freq_parameters()

    # added by @hassan
    if result_container:
        if check_should_cancel(json_file_path):
            result_container["results"][0]["responses"][0]["parameters"]["edt"] = []
            result_container["results"][0]["responses"][0]["parameters"]["t20"] = []
            result_container["results"][0]["responses"][0]["parameters"]["t30"] = []
            result_container["results"][0]["responses"][0]["parameters"]["c80"] = []
            result_container["results"][0]["responses"][0]["parameters"]["d50"] = []
            result_container["results"][0]["responses"][0]["parameters"]["ts"] = []
            result_container["results"][0]["responses"][0]["parameters"][
                "spl_t0_freq"
            ] = []

            result_container["results"][0]["responses"][0]["receiverResults"] = []

        else:
            result_container["results"][0]["responses"][0]["parameters"][
                "edt"
            ] = edt_band.tolist()
            result_container["results"][0]["responses"][0]["parameters"][
                "t20"
            ] = t20_band.tolist()
            result_container["results"][0]["responses"][0]["parameters"][
                "t30"
            ] = t30_band.tolist()
            result_container["results"][0]["responses"][0]["parameters"][
                "c80"
            ] = c80_band.tolist()
            result_container["results"][0]["responses"][0]["parameters"][
                "d50"
            ] = d50_band.tolist()
            result_container["results"][0]["responses"][0]["parameters"][
                "ts"
            ] = ts_band.tolist()
            result_container["results"][0]["responses"][0]["parameters"][
                "spl_t0_freq"
            ] = spl_r_t0_band.tolist()

            # result_container['frequenci[0]es'] = center_freq

            df = pd.DataFrame()
            for index, (edc_detail, pressure_detail) in enumerate(
                zip(spl_r_off_band, p_rec_off_deriv_band)
            ):
                result_container["results"][0]["responses"][0][
                    "receiverResults"
                ].append(
                    {
                        "data": edc_detail.tolist(),
                        # "data_pressure": pressure_detail.tolist(),
                        "t": (t_off - t_off[0]).tolist(),
                        "frequency": result_container["results"][0]["frequencies"][
                            index
                        ],
                        "type": "edc",
                    }
                )
                if "t" not in df.columns:
                    df["t"] = (t_off - t_off[0]).tolist()
                df[str(result_container["results"][0]["frequencies"][index]) + "Hz"] = (
                    pressure_detail.tolist()
                )

        with open(json_file_path, "w") as new_result_json:
            new_result_json.write(json.dumps(result_container, indent=4))
        with open(
            json_file_path.replace(".json", "_pressure.csv"), "w", newline=""
        ) as pressure_result_csv:
            df.to_csv(pressure_result_csv, index=False)

    et = time.time()  # end time
    elapsed_time = et - st

    # %%
    ###############################################################################
    # SAVING
    ###############################################################################
    # np.save(os.path.join('dt'),dt)
    # np.save(os.path.join('w_rec_off_band'),w_rec_off_band)
    # np.save(os.path.join('w_rec_off_deriv_band'),w_rec_off_deriv_band)
    # np.save(os.path.join('p_rec_off_deriv_band'),p_rec_off_deriv_band)
    # np.save(os.path.join('t_off'),t_off)

    # Save all variables to a file
    def save(filename):
        with open(filename, "wb") as f:
            # Filter out modules, functions, and other unsupported types
            filtered_variables = {}
            for k, v in globals().items():
                try:
                    # Check if the object can be pickled
                    pickle.dumps(v)
                    # Exclude some types explicitly known to cause issues
                    if not k.startswith("__") and not isinstance(
                        v,
                        (
                            types.ModuleType,
                            types.FunctionType,
                            types.BuiltinFunctionType,
                            types.LambdaType,
                            types.MethodType,
                            types.MappingProxyType,
                        ),
                    ):
                        filtered_variables[k] = v
                except Exception as e:
                    print(f"Could not pickle {k}: {str(e)}")

            pickle.dump(filtered_variables, f)

    # To save all current variables
    save("results.pkl")
    gmsh.finalize()


if __name__ == "__main__":
    from simulation_backend import (
        find_input_file_in_subfolders,
        create_tmp_from_input,
        save_results,
    )

    # Load the input file
    file_name = find_input_file_in_subfolders(
        os.path.dirname(__file__), "exampleInput_DE.json"
    )
    json_tmp_file = create_tmp_from_input(file_name)

    # Run the method
    de_method(json_tmp_file)

    # Save the results to a separate file
    save_results(json_tmp_file)



"""
Helper functions for room geometry checking and updating.
Contributor:
Anjana
Zeyu Xu
"""

import json
import os

import gmsh
import numpy as np
from scipy.spatial import ConvexHull

# -------------------------------
# Helper functions
# -------------------------------


def _open_gmsh_model(filename, mesh_dim=3):
    initialized_here = False
    if not gmsh.isInitialized():
        gmsh.initialize()
        initialized_here = True
    else:
        gmsh.clear()

    gmsh.open(filename)
    gmsh.model.mesh.generate(mesh_dim)
    return initialized_here


def _close_gmsh_model(initialized_here):
    gmsh.clear()
    if initialized_here:
        gmsh.finalize()


def get_points_and_tets():
    node_tags, coords, _ = gmsh.model.mesh.getNodes()
    points = np.array(coords).reshape(-1, 3)

    elem_types, elem_tags, elem_node_tags = gmsh.model.mesh.getElements(dim=3)
    if len(elem_node_tags) == 0:
        raise ValueError("No 3D elements (tetrahedra) found in mesh.")

    tets = np.array(elem_node_tags[0]).reshape(-1, 4) - 1  # zero-based
    return points, tets


def mesh_volume(points, tets):
    vol = 0.0
    for t in tets:
        p0, p1, p2, p3 = points[t]
        vol += abs(np.dot(np.cross(p1 - p0, p2 - p0), p3 - p0)) / 6.0
    return vol


def is_convex(points, tets, tol=1e-2):
    hull = ConvexHull(points)
    hull_vol = hull.volume
    mesh_vol = mesh_volume(points, tets)

    # If hull is much larger → non-convex (like L-shape)
    if (hull_vol - mesh_vol) / hull_vol > tol:
        return False
    return True


def get_corner_points():
    entities = sorted(gmsh.model.getEntities(dim=0), key=lambda entity: entity[1])
    points = []
    for dim, tag in entities:
        coords = gmsh.model.getValue(dim, tag, [])
        points.append(coords)
    return np.array(points).reshape(-1, 3)


def is_shoebox_corners(corners, tol=1e-6):
    x_vals = np.unique(np.round(corners[:, 0] / tol) * tol)
    y_vals = np.unique(np.round(corners[:, 1] / tol) * tol)
    z_vals = np.unique(np.round(corners[:, 2] / tol) * tol)
    return len(x_vals) == 2 and len(y_vals) == 2 and len(z_vals) == 2


def _get_node_coordinates():
    node_tags, coords, _ = gmsh.model.mesh.getNodes()
    coords = np.array(coords).reshape(-1, 3)
    return {int(tag): coords[index] for index, tag in enumerate(node_tags)}


def _get_triangles_for_entity(entity_tag, node_coordinates):
    elem_types, _, elem_node_tags = gmsh.model.mesh.getElements(2, entity_tag)
    triangles = []

    for elem_type, flat_node_tags in zip(elem_types, elem_node_tags):
        (
            element_name,
            _,
            _,
            nodes_per_element,
            _,
            _,
        ) = gmsh.model.mesh.getElementProperties(elem_type)
        element_nodes = np.array(flat_node_tags, dtype=int).reshape(
            -1, nodes_per_element
        )
        lower_name = element_name.lower()

        if "triangle" in lower_name:
            corner_sets = element_nodes[:, :3]
            triangles.extend(
                [
                    [
                        node_coordinates[int(node_set[0])],
                        node_coordinates[int(node_set[1])],
                        node_coordinates[int(node_set[2])],
                    ]
                    for node_set in corner_sets
                ]
            )
        elif "quadrangle" in lower_name:
            corner_sets = element_nodes[:, :4]
            for node_set in corner_sets:
                triangles.append(
                    [
                        node_coordinates[int(node_set[0])],
                        node_coordinates[int(node_set[1])],
                        node_coordinates[int(node_set[2])],
                    ]
                )
                triangles.append(
                    [
                        node_coordinates[int(node_set[0])],
                        node_coordinates[int(node_set[2])],
                        node_coordinates[int(node_set[3])],
                    ]
                )
        else:
            raise ValueError(
                f"Unsupported surface element type '{element_name}' for entity {entity_tag}."
            )

    if not triangles:
        return np.empty((0, 3, 3), dtype=float)
    return np.array(triangles, dtype=float)


def _get_face_vertices_for_entity(entity_tag):
    boundary_entities = gmsh.model.getBoundary(
        [(2, entity_tag)], oriented=False, recursive=True
    )
    point_tags = sorted({tag for dim, tag in boundary_entities if dim == 0})
    face_points = [gmsh.model.getValue(0, tag, []) for tag in point_tags]
    if not face_points:
        return np.empty((0, 3), dtype=float)
    return np.unique(np.array(face_points, dtype=float), axis=0)


def _compute_surface_area(triangles):
    if len(triangles) == 0:
        return 0.0

    v0 = triangles[:, 0, :]
    v1 = triangles[:, 1, :]
    v2 = triangles[:, 2, :]
    tri_areas = 0.5 * np.linalg.norm(np.cross(v1 - v0, v2 - v0), axis=1)
    return float(np.sum(tri_areas))


def _collect_surface_metrics():
    node_coordinates = _get_node_coordinates()
    room_areas = {}
    wall_centers = {}

    for dim, physical_tag in sorted(
        gmsh.model.getPhysicalGroups(dim=2), key=lambda item: item[1]
    ):
        group_name = (
            gmsh.model.getPhysicalName(dim, physical_tag) or f"Surface_{physical_tag}"
        )
        entity_tags = gmsh.model.getEntitiesForPhysicalGroup(dim, physical_tag)

        triangles = []
        face_points = []
        for entity_tag in entity_tags:
            entity_triangles = _get_triangles_for_entity(entity_tag, node_coordinates)
            if len(entity_triangles) > 0:
                triangles.extend(entity_triangles.tolist())
            entity_face_points = _get_face_vertices_for_entity(entity_tag)
            if len(entity_face_points) > 0:
                face_points.extend(entity_face_points.tolist())

        area = _compute_surface_area(np.array(triangles, dtype=float))
        if face_points:
            face_center = np.mean(
                np.unique(np.array(face_points, dtype=float), axis=0), axis=0
            )
        else:
            face_center = np.zeros(3)
        room_areas[group_name] = area
        wall_centers[group_name] = np.round(face_center, 4).tolist()

    return room_areas, wall_centers


def _ensure_geometry_section(data):
    if "geometry" not in data:
        data["geometry"] = [{}]
    elif len(data["geometry"]) == 0:
        data["geometry"].append({})
    return data["geometry"][0]


def _write_geometry_fields(json_file_path, **fields):
    if not os.path.exists(json_file_path):
        raise FileNotFoundError(f"JSON file not found: {json_file_path}")

    with open(json_file_path, "r") as file_obj:
        data = json.load(file_obj)

    geometry = _ensure_geometry_section(data)
    geometry.update(fields)

    with open(json_file_path, "w") as file_obj:
        json.dump(data, file_obj, indent=4)


def collect_room_geometry_data(geo_file):
    if not os.path.exists(geo_file):
        raise FileNotFoundError(f"GEO file not found: {geo_file}")

    initialized_here = _open_gmsh_model(geo_file, mesh_dim=3)

    try:
        points, tets = get_points_and_tets()
        volume = mesh_volume(points, tets)

        convex = is_convex(points, tets)
        corners = get_corner_points()
        shoebox = is_shoebox_corners(corners)

        if convex:
            room = "convex"
        else:
            room = "shoebox"

        room_areas, wall_centers = _collect_surface_metrics()

        return {
            "vertices": corners.tolist(),
            "wall_centers": wall_centers,
            "room_areas": room_areas,
            "room_volumn": volume,
            "room": room,
            "shoebox": shoebox,
        }
    finally:
        _close_gmsh_model(initialized_here)


# -------------------------------
# Main
# -------------------------------
def get_room_geometry(geo_file):
    geometry_data = collect_room_geometry_data(geo_file)
    return geometry_data["room_volumn"], geometry_data["room"]


def sync_room_geometry(json_file_path, geo_file_path):
    geometry_data = collect_room_geometry_data(geo_file_path)
    _write_geometry_fields(
        json_file_path,
        vertices=geometry_data["vertices"],
        wall_centers=geometry_data["wall_centers"],
        room_areas=geometry_data["room_areas"],
        room_volumn=geometry_data["room_volumn"],
    )
    return geometry_data["room_volumn"], geometry_data["room"]


def update_surface_areas(json_file_path, geo_file_path):
    """
    Calculate surface areas from the Gmsh .geo file and write them
    into the JSON under "geometry[0]['room_areas']".
    """
    geometry_data = collect_room_geometry_data(geo_file_path)
    areas = geometry_data["room_areas"]
    _write_geometry_fields(json_file_path, room_areas=areas)

    print("✅ Surface areas updated in JSON:")
    for name, area in areas.items():
        print(f"  {name}: {area:.2f} m²")

    # return areas


def update_wall_centers(json_path, geo_path):
    geometry_data = collect_room_geometry_data(geo_path)
    _write_geometry_fields(
        json_path,
        vertices=geometry_data["vertices"],
        wall_centers=geometry_data["wall_centers"],
    )

    print("✅ Wall centers updated successfully!")



import time
import os
import sys
import psutil
import numpy as np
import ray
import scipy.special as scy
import matplotlib.pyplot as plt

from deism.core_deism import (
    pre_calc_Wigner,
    get_directivity_coefs,
    ray_run_DEISM_ARG_ORG,
    init_source_directivities_ARG,
    init_receiver_directivities_ARG,
    vectorize_C_vu_r,
    vectorize_C_nm_s_ARG,
    ray_run_DEISM_ARG_LC_matrix,
    ray_run_DEISM_ARG_MIX,
)

from deism.core_deism_arg import (
    Room_deism_cpp,
    get_ref_paths_ARG,
    rotate_room_src_rec,
    find_wall_centers,
)
from deism.data_loader import (
    cmdArgsToDict_ARG,
    printDict,
    detect_conflicts,
)
from deism.utilities import plot_RTFs


def init_parameters_convex(params):
    """
    Initialize some important additional parameters for DEISM-ARG
    In addition to the ones defined in configSingleParam_ARG.yaml
    """
    # Room vertices
    vertices = np.array(
        [
            [0, 0, 0],  # Origin
            [0, 0, 3.5],
            [0, 3, 2.5],
            [0, 3, 0],
            [4, 0, 0],
            [4, 0, 3.5],
            [4, 3, 2.5],
            [4, 3, 0],
        ]
    )
    # If you need to find the wall centers, use the following code
    # This may be useful for some applications, e.g., bonding the impedance to the walls using the room center
    # wall_centers = find_wall_centers(vertices)
    wallCenters = np.array(
        [
            [0, 1.5, 1.5],
            [2, 3, 1.25],
            [4, 1.5, 1.5],
            [2, 0, 1.75],
            [2, 1.5, 0],
            [2, 1.5, 3],
        ]
    )
    # You need to adlign the acoustic impedance to the wall centers defined above
    # The format of the impedance is [number of walls, len(freqs)], 2D numpy array
    # params["acousImpend"] = np.arange(1, 6 * len(params["freqs"]) + 1).reshape(
    #     6, len(params["freqs"])
    # )
    params["acousImpend"] = (
        np.ones((6, len(params["freqs"]))) * 18
    )  # recreating the old example
    # --- Room rotation, if rotate the room w.r.t the origin ---
    if_rotate_room = 1
    # --- Room rotation angles using Z-X-Z Euler angles ---
    # [alpha, beta, gamma], 3D Euler angles, The rotation matrix calculation used in COMSOL, see:
    # https://doc.comsol.com/5.5/doc/com.comsol.help.comsol/comsol_ref_definitions.12.092.html
    # The chosen rotation angles are used to illustrate a scenario where we can compare with the room created using pyroomacoustics
    # See file DEISM_ARG_comparisons.py
    room_rotation = np.array([90, 90, 90])  # [alpha, beta, gamma] in degrees
    # --- Add the above parameters to the params dictionary ---
    params["vertices"] = vertices
    params["wallCenters"] = wallCenters
    params["if_rotate_room"] = if_rotate_room
    params["room_rotation"] = room_rotation
    # Apply room rotation to the room vertices and source/receiver positions
    if if_rotate_room:
        params = rotate_room_src_rec(params)
    return params


def main():
    params, cmdArgs = cmdArgsToDict_ARG("configSingleParam_arg.yml")
    params = init_parameters_convex(params)
    # detect conflicts
    detect_conflicts(params)
    # print the parameters or not
    if cmdArgs.quiet:
        params["silentMode"] = 1
    # print the parameters
    printDict(params)
    # -------------------------------------------------------
    # If you want to modify the parameters, you can also do it here
    # by changing the values of the parameters in the params dictionary
    # -------------------------------------------------------
    # Save only the current parameters
    params_save = params.copy()
    # If run DEISM function, run with changed parameters if --run flag is set in the cmd
    # If cmdArgs are all None values, run following codes directily
    if cmdArgs.run or all(
        value in [None, False] for value in vars(cmdArgs).values()
    ):  # no input in cmd will also run using default parameters
        # Initialize directivities
        # -------------------------------------------------------
        # --- Loading and calculating receiver directivities ----
        # -------------------------------------------------------
        params = init_receiver_directivities_ARG(
            params, params["if_rotate_room"], room_rotation=params["room_rotation"]
        )
        # -----------------------------------------------------------
        # -- Using DESIM-ARG with c++ libroom_deism to find
        # the visible images and reflection matrices, attenuations --
        # -----------------------------------------------------------
        # initialize Room_deism
        room_pra_deism = Room_deism_cpp(params)
        # get the reflection paths used for the DEISM-ARG calculation
        params = get_ref_paths_ARG(params, room_pra_deism)
        # ---------------------------------------------------------------
        # -- Calculating the reflected source directivity coefficients --
        # ---------------------------------------------------------------
        params = init_source_directivities_ARG(
            params,
            params["if_rotate_room"],
            params["reflection_matrix"],
            room_rotation=params["room_rotation"],
        )
        # -------------------------------------------------------
        # ------ Testing DEISM-ARG using LC version -------------
        # -------------------------------------------------------
        params = vectorize_C_vu_r(params)
        params = vectorize_C_nm_s_ARG(params)

        # Initialize Ray
        num_cpus = psutil.cpu_count(logical=False)
        if not ray.is_initialized():
            ray.init(num_cpus=num_cpus, log_to_driver=False)
            print("\n")
        # run DEISM-ARG using LC versionyaok
        P_DEISM_ARG_LC = ray_run_DEISM_ARG_LC_matrix(params, params["images"])
        # -------------------------------------------------------
        # -- Calculate Wigner 3J matrices for some DEISM modes --
        # -------------------------------------------------------
        if params["DEISM_mode"] == "ORG" or params["DEISM_mode"] == "MIX":
            params["Wigner"] = pre_calc_Wigner(params)
        # ------------------------------------------------------------------------
        # ------ Run DEISM-ARG using the original version and mixed version ------
        # ------------------------------------------------------------------------
        # Run DEISM
        P_DEISM_ARG = ray_run_DEISM_ARG_ORG(params, params["images"], params["Wigner"])
        # Run DEISM-ARG with mixed version
        P_DEISM_ARG_MIX = ray_run_DEISM_ARG_MIX(
            params, params["images"], params["Wigner"]
        )
        # Shutdown Ray
        ray.shutdown()
        # -------------------------------------------------------
        # -- Plot the RTFs for the three versions of DEISM-ARG --
        # -------------------------------------------------------
        # We now also save some parameters in the name of the figure
        figure_name = "DEISM_ARGs_compare_RO_{}_{}_vertices_src_{:.2f}_{:.2f}_{:.2f}_rec_{:.2f}_{:.2f}_{:.2f}".format(
            params["maxReflOrder"],
            len(params["vertices"]),
            params["posSource"][0],
            params["posSource"][1],
            params["posSource"][2],
            params["posReceiver"][0],
            params["posReceiver"][1],
            params["posReceiver"][2],
        )
        # Define the save path for the figures
        save_path = "./outputs/figures"
        # check if the save path exists
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        # Define the parameters for the plot
        # P_all is a list of the RTFs for the three versions of DEISM-ARG
        P_all = [P_DEISM_ARG, P_DEISM_ARG_LC, P_DEISM_ARG_MIX]
        # P_labels is a list of the labels for the three versions of DEISM-ARG
        P_labels = ["DEISM-ARG", "DEISM-ARG LC", "DEISM-ARG MIX"]
        # P_freqs is a list of the frequency bins for the RTFs
        P_freqs = [params["freqs"], params["freqs"], params["freqs"]]
        # Define other parameters for the plot
        PLOT_SCALE = (
            "dB"  # can be "dB" or "linear", the scale of the sound pressure levels
        )
        IF_FREQS_Log = 1  # if the frequency bins are plotted in log scale
        IF_SAME_MAGSCALE = 0  # if the magnitude is plotted in a fixed scale
        IF_UNWRAP_PHASE = 0  # if the phase is unwrapped
        IF_SAVE_PLOT = 1  # if the plot is saved
        # Plot the RTFs
        plot_RTFs(
            figure_name,
            save_path,
            P_all,
            P_labels,
            P_freqs,
            PLOT_SCALE,
            IF_FREQS_Log,
            IF_SAME_MAGSCALE,
            IF_UNWRAP_PHASE,
            IF_SAVE_PLOT,
        )

        # -------------------------------------------------------
        # Save the results to local directory with .npz format
        save_path = "./outputs/RTFs"
        # check if the save path exists
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        # Save the results along with all the parameters to a .npz file with file name as the current time
        # Save only the initial parameters which are used to run the DEISM function
        np.savez(
            f"{save_path}/DEISM_ARG_RTF_{time.strftime('%Y%m%d_%H%M%S')}",
            P_DEISM=P_DEISM_ARG,
            params=params_save,
        )
    else:
        print("DEISM Function has not been run because --run flag was not set.\n")


# -------------------------------------------------------
if __name__ == "__main__":
    main()



import numpy as np
from deism.core_deism import *
from deism.data_loader import *


def test_deism_arg_simulation():
    """Test DEISM-ARG simulation with impedance update."""
    deism = DEISM("RIR", "convex")
    # Testing impedance update
    # Example of room volumn and roomAreas
    roomVolumn = 36
    roomAreas = np.array([9, 10, 9, 10, 12, np.sqrt(10) * 4])
    deism.update_room(roomVolumn=roomVolumn, roomAreas=roomAreas)
    deism.update_wall_materials()
    deism.update_wall_materials(
        np.array([[100, 100, 100, 100, 100, 100], [100, 100, 100, 100, 100, 100]]).T,
        np.array([10, 20]),
        "impedance",
    )
    deism.update_freqs()
    deism.update_source_receiver()
    deism.update_directivities()
    deism.run_DEISM()
    pressure = deism.params["RTF"]

    # Basic assertions to verify the simulation ran successfully
    assert pressure is not None, "Pressure should not be None"
    assert len(pressure) > 0, "Pressure should have data"
    assert deism.params["posSource"] is not None, "Source position should be set"
    assert deism.params["posReceiver"] is not None, "Receiver position should be set"
    assert deism.params["freqs"] is not None, "Frequencies should be set"

    print("Simulation done!")


# -------------------------------------------------------
if __name__ == "__main__":
    test_deism_arg_simulation()



# This function originated from https://github.com/LCAV/pyroomacoustics and is modified by Fraunhofer
# The code was obtained under the MIT license, which is distributed with this software
# This configuration file is used to install the DEISM-ARG package.
# Is was modified from the setup.py file of the pyroomacoustics package.
# Link to the original pyroomacoustics setup.py file:
# https://github.com/LCAV/pyroomacoustics/tree/master/setup.py


#!/usr/bin/env python
from __future__ import print_function

import os
import sys

# To use a consistent encoding
from os import path
import shutil
import platform

# Read version
with open("deism/version.py") as f:
    exec(f.read())
try:
    from setuptools import Extension, distutils, setup, find_packages, Command
    from setuptools.command.build_ext import build_ext
    from setuptools.command.install_lib import install_lib
except ImportError:
    print("Setuptools unavailable. Falling back to distutils.")
    import distutils
    from distutils.command.build_ext import build_ext
    from distutils.core import setup
    from distutils.extension import Extension


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked."""

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


# build C extension for image source model
# libroom_src_dir = "pyroomacoustics/libroom_src"
libroom_src_dir = "./deism/libroom_src/"
libroom_files = [
    os.path.join(libroom_src_dir, f)
    for f in [
        "room.hpp",
        "room.cpp",
        "wall.hpp",
        "wall.cpp",
        "microphone.hpp",
        "geometry.hpp",
        "geometry.cpp",
        "common.hpp",
        "rir_builder.cpp",
        "rir_builder.hpp",
        "libroom.cpp",
        "threadpool.hpp",
    ]
]
# Set extra compile arguments conditionally
extra_compile_args = ["-DEIGEN_MPL2_ONLY", "-Wall", "-O3", "-DEIGEN_NO_DEBUG"]
extra_link_args = []
# Only add "-arch arm64" if running on ARM-based macOS
if sys.platform == "darwin" and platform.machine() == "arm64":
    extra_compile_args += ["-arch", "arm64"]
    extra_link_args = ["-arch", "arm64"]
# When you run python setup.py build_ext --inplace or pip install .,
# the Extension specified in the ext_modules list will trigger the build process.
ext_modules = [  # This specifies the C++ extension that will be built.
    Extension(
        "deism.libroom_deism",  # The name of the module deism.libroom_deism.
        [
            os.path.join(libroom_src_dir, f)
            for f in [
                "libroom.cpp",
                "rir_builder.cpp",
            ]  # The .cpp files that need to be compiled.
        ],  #
        depends=libroom_files,  # The C++ files that the extension depends on.
        include_dirs=[  # Directories that contain header files needed for compilation
            ".",
            libroom_src_dir,
            str(get_pybind_include()),
            str(get_pybind_include(user=True)),
            os.path.join(libroom_src_dir, "ext/eigen"),
        ],
        language="c++",
        extra_compile_args=extra_compile_args,  # Additional compilation flags
        extra_link_args=extra_link_args,
    ),
    # Extension(
    #     "pyroomacoustics.build_rir",
    #     # ["pyroomacoustics/build_rir.pyx"],
    #     ["libroom_src/build_rir.pyx"],
    #     language="c",
    #     extra_compile_args=[],
    # ),
]

# here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
# with open(path.join(here, "README.rst"), encoding="utf-8") as f:
#     long_description = f.read()


### Build Tools (taken from pybind11 example) ###


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, "-std=c++14"):
        return "-std=c++14"
    elif has_flag(compiler, "-std=c++11"):
        return "-std=c++11"
    else:
        raise RuntimeError(
            "Unsupported compiler -- at least C++11 support " "is needed!"
        )


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {
        "msvc": ["/EHsc"],
        "unix": [],
    }

    if sys.platform == "darwin":
        c_opts["unix"] += ["-stdlib=libc++", "-mmacosx-version-min=10.7"]

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == "unix":
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            if ext.language == "c++":
                ext.extra_compile_args += opts
                ext.extra_link_args += opts
        build_ext.build_extensions(self)

        # Compile count_reflections.cpp as a standalone shared library
        self.build_count_reflections_lib()

        # Add the following two lines to remove the build directory after building
        self.clean()

    def build_count_reflections_lib(self):
        """Compile count_reflections.cpp as a standalone shared library."""
        deism_dir = os.path.join(os.path.dirname(__file__), "deism")
        cpp_file = os.path.join(deism_dir, "count_reflections.cpp")

        if not os.path.exists(cpp_file):
            print(
                f"Warning: {cpp_file} not found, skipping count_reflections compilation"
            )
            return

        # Determine library extension based on platform
        if sys.platform == "darwin":
            lib_ext = ".dylib"
        elif sys.platform == "win32":
            lib_ext = ".dll"
        else:
            lib_ext = ".so"

        # Compile into source directory (for editable installs)
        lib_file_source = os.path.join(deism_dir, f"count_reflections{lib_ext}")

        # Also compile into build_lib if it exists (for regular installs/wheels)
        build_lib = getattr(self, "build_lib", None)
        if build_lib:
            lib_file_build = os.path.join(
                build_lib, "deism", f"count_reflections{lib_ext}"
            )
            # Ensure the directory exists
            os.makedirs(os.path.dirname(lib_file_build), exist_ok=True)
        else:
            lib_file_build = None

        # Compile command
        if sys.platform == "win32":
            # Windows: use g++ if available
            compile_cmd = [
                "g++",
                "-shared",
                "-fPIC",
                "-O3",
                "-std=c++11",
                cpp_file,
                "-o",
                lib_file_source,
            ]
        else:
            # Unix-like (Linux, macOS)
            compile_cmd = [
                "g++",
                "-shared",
                "-fPIC",
                "-O3",
                "-std=c++11",
                cpp_file,
                "-o",
                lib_file_source,
            ]

        print(f"Compiling count_reflections library: {' '.join(compile_cmd)}")
        try:
            import subprocess

            result = subprocess.run(
                compile_cmd, check=True, capture_output=True, text=True
            )
            print(f"Successfully compiled count_reflections{lib_ext}")

            # If build_lib exists, copy the compiled library there too
            if lib_file_build and os.path.exists(lib_file_source):
                shutil.copy2(lib_file_source, lib_file_build)
                print(f"Copied library to build directory: {lib_file_build}")
        except subprocess.CalledProcessError as e:
            print(f"Warning: Failed to compile count_reflections library:")
            print(f"  Command: {' '.join(compile_cmd)}")
            print(f"  Error: {e.stderr}")
            print(
                "  The package will still work, but count_reflections_cpp will not be available."
            )
        except FileNotFoundError:
            print(
                "Warning: g++ compiler not found. count_reflections library will not be compiled."
            )
            print(
                "  The package will still work, but count_reflections_cpp will not be available."
            )

    def clean(self):
        """Custom clean step after build is complete."""
        temp_folders = ["var", "tmp"]
        for folder in temp_folders:
            tmp_folder = os.path.join(os.getcwd(), folder)
            if os.path.exists(tmp_folder) and os.path.isdir(tmp_folder):
                shutil.rmtree(tmp_folder)
                print(f"Temporary folder {tmp_folder} has been removed!")


#     def clean_build_directory(self):
#         """Remove the build folder after installation is complete"""
#         build_folder = os.path.join(os.getcwd(), "build")
#         if os.path.exists(build_folder) and os.path.isdir(build_folder):
#             shutil.rmtree(build_folder)
#             print(f"Build folder {build_folder} has been removed!")

#         # Optionally remove other temp folders
#         temp_folders = ["var", "tmp"]
#         for folder in temp_folders:
#             tmp_folder = os.path.join(os.getcwd(), folder)
#             if os.path.exists(tmp_folder) and os.path.isdir(tmp_folder):
#                 shutil.rmtree(tmp_folder)
#                 print(f"Temporary folder {tmp_folder} has been removed!")


### Build Tools End ###


setup_kwargs = dict(
    name="deism",
    version=__version__,
    # packages=find_packages(),
    packages=["deism"],
    description="An image source-based method used to simulate room transfer functions for arbitrary room shapes.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Zeyu Xu",
    author_email="zeyu.xu@audiolabs-erlangen.de",
    url="https://github.com/audiolabs/DEISM",
    # license="MIT", !!! Todo: add license
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # Libroom C extension
    ext_modules=ext_modules,
    # Necessary to keep the source files
    # package_data={"DEISM": ["*.pxd", "*.pyx", "data/materials.json"]},
    # here controls where the pyd shared lib will be copied.
    # package_dir={"": os.path.join(os.getcwd(), "deism")},
    python_requires=">=3.9",
    install_requires=[
        "numpy",
        "scipy",
        "sympy",
        "psutil",
        "matplotlib",
        "ray",
        "sound-field-analysis",
        "pybind11>=2.2",
        "Cython",
        "pyroomacoustics",
    ],
    extras_require={
        "geometry": ["gmsh"],
    },
    cmdclass={
        "build_ext": BuildExt,
    },  # taken from pybind11 example
)

setup(**setup_kwargs)



import os
import time
import numpy as np
import json
import logging
import traceback

# from Diffusion.acousticDE.FiniteVolumeMethod.FVMfunctions import create_vgroups_names
import gmsh
from deism.core_deism import *
from deism.data_loader import *
from deism.room_check import sync_room_geometry


logger = logging.getLogger(__name__)


def create_example_tmp_input(
    input_file_name="exampleInput_Deism.json", tmp_file_name="exampletmp_deism.json"
):
    """Create a temporary copy of the example input JSON."""
    examples_dir = os.path.dirname(os.path.abspath(__file__))
    input_path = os.path.join(examples_dir, input_file_name)
    if not os.path.exists(input_path):
        raise FileNotFoundError(f"Example input file not found: {input_path}")

    with open(input_path, "r", encoding="utf-8") as json_input:
        data = json.load(json_input)

    output_dir = os.path.join(examples_dir, "output")
    os.makedirs(output_dir, exist_ok=True)
    json_tmp_file = os.path.join(output_dir, tmp_file_name)

    with open(json_tmp_file, "w", encoding="utf-8") as json_output:
        json.dump(data, json_output, indent=4)

    return json_tmp_file


def create_vgroups_names(file_path):
    """
    Create a list of the material names assigned in SketchUp

    Parameters
    ----------
        file_path : str
            Full path to the mesh file

    Returns
    -------
        vGroupsNames : list
            Names of the materials in the msh file (the material name are the same as the one assigned in the SketchUp file)
    """
    gmsh.initialize()  # Initialize msh file
    mesh = gmsh.open(file_path)  # open the file
    dim = (
        -1
    )  # dimensions of the entities, 0 for points, 1 for curves/edge/lines, 2 for surfaces, 3 for volumes, -1 for all the entities
    tag = -1  # all the nodes of the room
    vGroups = gmsh.model.getPhysicalGroups(
        -1
    )  # these are the entity tag and physical groups in the msh file.
    vGroupsNames = (
        []
    )  # these are the entity tag and physical groups in the msh file + their names
    for iGroup in vGroups:
        dimGroup = iGroup[
            0
        ]  # entity tag: 1 lines, 2 surfaces, 3 volumes (1D, 2D or 3D)
        tagGroup = iGroup[
            1
        ]  # physical tag group (depending on material properties defined in SketchUp)
        namGroup = gmsh.model.getPhysicalName(
            dimGroup, tagGroup
        )  # names of the physical groups defined in SketchUp
        alist = [
            dimGroup,
            tagGroup,
            namGroup,
        ]  # creates a list of the entity tag, physical tag group and name
        # print(alist)
        vGroupsNames.append(alist)

    return vGroupsNames


def parse_value(val):
    """Handle strings of comma-separated floats OR single float values."""
    if isinstance(val, str):
        return np.array([float(x.strip()) for x in val.split(",") if x.strip()])
    elif isinstance(val, (int, float)):
        return np.array([val])
    elif isinstance(val, (list, tuple)):
        return np.array(val, dtype=float)
    else:
        raise ValueError(f"Unsupported type for parse_value: {type(val)}")


def get_deism_surface_order(vgroups_names):
    """Map Gmsh physical surfaces to DEISM's expected wall order."""
    surface_names_by_tag = {
        int(tag): name for dim, tag, name in vgroups_names if int(dim) == 2
    }
    deism_tag_order = [2, 5, 4, 6, 1, 3]
    missing_tags = [tag for tag in deism_tag_order if tag not in surface_names_by_tag]
    if missing_tags:
        raise KeyError(
            f"Missing physical surface tags required by DEISM: {missing_tags}"
        )
    return [surface_names_by_tag[tag] for tag in deism_tag_order]


def deism_method(json_file_path=None):
    """
    DEISM simulation method that processes a JSON file containing simulation parameters.

    Parameters
    ----------
    json_file_path : str, optional
        Path to the JSON file containing simulation parameters and results.
        If None, the method will not run.
    """
    print("deism_method: starting simulation")
    st = time.time()  # start time of calculation

    if json_file_path is None:
        print("No JSON file path provided. Exiting.")
        return

    # Change to the directory containing the config file
    script_dir = os.path.dirname(os.path.abspath(__file__))
    original_cwd = os.getcwd()
    os.chdir(script_dir)

    try:
        # Step 1: read JSON to get geo_path
        with open(json_file_path, "r", encoding="utf-8") as json_file:
            result_container = json.load(json_file)
            geo_path = result_container["geo_path"]

        # Step 2: update areas, wall centers, vertices, and volume in one pass
        _, room = sync_room_geometry(json_file_path, geo_path)

        with open(json_file_path, "r", encoding="utf-8") as json_file:
            result_container = json.load(json_file)

        vGroupsNames = create_vgroups_names(result_container["geo_path"])
        print("vGroupsNames", vGroupsNames)

        # Checking whether the 'should_cancel' flag has been set to True by the user
        # Do not call this function all the time, as it is quite heavy
        # This function should be called in the main calculation loop
        def check_should_cancel(json_file_path_in):
            try:
                if json_file_path_in is not None:
                    with open(json_file_path_in, "r") as json_file_to_check:
                        data = json.load(json_file_to_check)
                        # Update the specified field value
                        if "should_cancel" in data:
                            return data["should_cancel"]
                return False
            except Exception as e:
                print("check_should_cancel returned: " + str(e))
                print(traceback.format_exc())
                return False

        if check_should_cancel(json_file_path):
            return

        # Load from the json file
        print("Obtaining simulation settings from the json file ... \n")
        simulation_settings = None
        coord_source = None
        coord_rec = None
        abs_coeffs_loaded = None
        freq_bands = None

        if result_container:
            simulation_settings = result_container["simulationSettings"]
            coord_source = [
                result_container["results"][0]["sourceX"],
                result_container["results"][0]["sourceY"],
                result_container["results"][0]["sourceZ"],
            ]

            coord_rec = [
                result_container["results"][0]["responses"][0]["x"],
                result_container["results"][0]["responses"][0]["y"],
                result_container["results"][0]["responses"][0]["z"],
            ]
            abs_coeffs_loaded = result_container["absorption_coefficients"]
            freq_bands = np.array(result_container["results"][0]["frequencies"])

        # Convert data to the ones needed in DEISM
        print("Converting data to the ones needed in DEISM ... \n")
        # -----------------------------------------------------------
        # About room geometry and wall properties
        # N is the number of vertices of the room
        # M is the number of wall centers
        # -----------------------------------------------------------
        vertices = np.array(
            result_container["geometry"][0]["vertices"]
        )  # Nx3 numpy array
        wall_centers_loaded = result_container["geometry"][0]["wall_centers"]
        room_volumn = result_container["geometry"][0]["room_volumn"]  # float
        room_areas_loaded = result_container["geometry"][0][
            "room_areas"
        ]  # (M,) numpy array
        # we want the absorption has size 6 * len(frequency bands)
        # The first dimension is for the walls, viz., x1, x2, y1, y2, z1, z2
        # Corresponding to wall 1, wall 3, wall 2, wall 4, floor, ceiling

        wall_order = get_deism_surface_order(vGroupsNames)
        # Create an empty array for the absorption coefficients
        absorption_coefficients = np.zeros((6, len(freq_bands)))
        wall_centers = np.zeros((6, 3))
        room_areas = np.zeros((6, 1))
        for index, wall in enumerate(wall_order):
            absorption_coefficients[index, :] = parse_value(abs_coeffs_loaded[wall])
            wall_centers[index, :] = parse_value(wall_centers_loaded[wall])
            room_areas[index, :] = parse_value(room_areas_loaded[wall])

        # Apply DEISM
        deism = DEISM("RIR", room)
        print("valuess of deism")  #
        print("vertices", vertices)
        print("wall center", wall_centers)
        print("room areas", room_areas)

        deism.update_room(
            vertices,
            wall_centers,
            room_volumn,
            room_areas,
        )
        deism.update_wall_materials(
            absorption_coefficients, freq_bands, "absorpCoefficient"
        )
        deism.update_freqs()
        # deism.update_images()
        # update source and receiver positions in deism
        deism.params["posSource"] = np.array(coord_source)
        deism.params["posReceiver"] = np.array(coord_rec)
        deism.update_source_receiver()
        deism.update_directivities()
        deism.run_DEISM()
        rir = deism.get_results()
        # -----------------------------------------------------------
        # Save the simulation results
        # -----------------------------------------------------------
        # Save the simulation results in the json file
        result_container["results"][0]["responses"][0]["receiverResults"] = rir.tolist()
        with open(json_file_path, "w") as new_result_json:
            new_result_json.write(json.dumps(result_container, indent=4))
        print("deism_method: simulation done!")

    except Exception as e:
        print(f"Error in deism_method: {str(e)}")
        print(traceback.format_exc())
    finally:
        # Restore original working directory
        os.chdir(original_cwd)


# -------------------------------------------------------
if __name__ == "__main__":
    import os
    import sys

    # Add the parent directory to Python path to allow importing simulation_backend
    current_dir = os.path.dirname(os.path.abspath(__file__))
    parent_dir = os.path.dirname(current_dir)
    if parent_dir not in sys.path:
        sys.path.insert(0, parent_dir)

    from headless_backend.HelperFunctions import save_results

    # Load the example input and resolve geometry assets from examples/data/geometry
    json_tmp_file = create_example_tmp_input(
        "exampleInput_Deism.json", "exampletmp_deism.json"
    )

    # Run the method
    deism_method(json_tmp_file)

    # Save the results to a separate file
    save_results(json_tmp_file)



import os
import numpy as np
import matplotlib.pyplot as plt
import numpy.typing as npt
from typing import Optional
from scipy.signal import lfilter, sosfiltfilt, tf2sos


def _compute_highpass_coeffs(fs: float, fcut: float) -> tuple[np.ndarray, np.ndarray]:
    """Return numerator and denominator coefficients for the RIR high-pass filter."""
    W = 2.0 * np.pi * float(fcut) / float(fs)
    R1 = np.exp(-W)
    B1 = 2.0 * R1 * np.cos(W)
    B2 = -R1 * R1
    A1 = -(1.0 + R1)

    b = np.array([1.0, A1, R1], dtype=np.float64)
    a = np.array([1.0, -B1, -B2], dtype=np.float64)
    return b, a


def highpass_RIR(
    h: npt.ArrayLike,
    fs: float,
    fcut: float = 100.0,
    *,
    axis: int = 0,
    zero_phase: bool = False,
    dtype: Optional[npt.DTypeLike] = None,
) -> np.ndarray:
    """Apply the audiolabs high-pass filter to room impulse responses.

    Parameters
    ----------
    h:
        Array containing the room impulse response. Time samples must be located along
        ``axis``.
    fs:
        Sampling rate in Hertz. Must be positive.
    fcut:
        High-pass cutoff frequency in Hertz. Must satisfy ``0 < fcut < fs/2``. Defaults
        to 100 Hz as used in ``audiolabs/rir-generator``.
    axis:
        Axis along which the filter is applied. Defaults to 0 and accepts negative
        indices.
    zero_phase:
        When ``True`` the filter is applied in a zero-phase manner using
        :func:`scipy.signal.sosfiltfilt`, avoiding phase distortion at the cost of an
        additional reverse pass. Defaults to ``False`` (causal :func:`scipy.signal.lfilter`).
    dtype:
        Optional NumPy dtype specification used for the internal representation and
        output. When ``None`` the result uses at least single precision while preserving
        higher precision inputs.

    Returns
    -------
    numpy.ndarray
        Filtered impulse response with the same shape as ``h``.

    Raises
    ------
    ValueError
        If ``fs`` is non-positive, ``fcut`` is outside ``(0, fs/2)``, or ``axis`` is
        incompatible with the input shape.
    """
    if fs <= 0:
        raise ValueError("fs must be positive.")
    if not (0.0 < fcut < fs / 2.0):
        raise ValueError(f"fcut must be in (0, fs/2). Got fcut={fcut}, fs={fs}.")

    target_dtype = dtype or np.result_type(h, np.float32)
    h_array = np.asarray(h, dtype=target_dtype)

    if h_array.ndim == 0:
        raise ValueError("x must be at least 1-dimensional.")

    if axis < 0:
        axis += h_array.ndim
    if axis < 0 or axis >= h_array.ndim:
        raise ValueError(
            f"axis={axis} is out of bounds for input with ndim={h_array.ndim}."
        )

    b, a = _compute_highpass_coeffs(fs, fcut)

    if zero_phase:
        sos = tf2sos(b, a)
        filtered = sosfiltfilt(sos, h_array, axis=axis)
    else:
        filtered = lfilter(b, a, h_array, axis=axis)

    return np.asarray(filtered, dtype=target_dtype)


def cart2sph(x, y, z):
    """Convert cartesian coordinates x, y, z to spherical coordinates az, el, r."""
    H_xy = np.hypot(x, y)
    r = np.hypot(H_xy, z)
    el = np.arctan2(z, H_xy)
    az = np.arctan2(y, x)

    return az, el, r


def sph2cart(az, el, r):
    """Convert spherical coordinates az, el, r to cartesian coordinates x, y, z."""
    rcos_theta = r * np.cos(el)
    x = rcos_theta * np.cos(az)
    y = rcos_theta * np.sin(az)
    z = r * np.sin(el)

    return x, y, z


def get_SPL(P):
    P_rms = np.abs(np.sqrt(0.5 * P * P.conjugate()))
    P0 = 20 * 10 ** (-6)
    SPL = 20 * np.log10(P_rms / P0)
    return SPL


# %% plot_results
def plot_RTFs(
    figure_name,
    save_path,
    P_all,
    P_labels,
    P_freqs,
    PLOT_SCALE="dB",
    IF_FREQS_Log=1,
    IF_SAME_MAGSCALE=0,
    IF_UNWRAP_PHASE=0,
    IF_SAVE_PLOT=1,
):
    """
    Plot the comparisons between two RTF simulations.
    Input:
    1. figure_name: the name ending of the figure, you can set any name you like
    2. save_path: the path to save the figure
    3. P_all: A list containing all RTFs to be plotted, each one is a numpy array
    4. P_labels: A list containing the labels for each RTF
    5. P_freqs: A list containing the frequency axis for each RTF
    6. PLOT_SCALE: 'dB' or 'linear', default is 'dB', plot the Sound pressure levels in dB or linear scale
    7. IF_FREQS_Log: 1 or 0, default is 1, plot the frequency in log scale or linear scale
    8. IF_SAME_MAGSCALE: 1 or 0, default is 0, plot the magnitude in a fixed scale or not, you can set the scale in the code "ax.set_ylim([0, 60])"
    9. IF_UNWRAP_PHASE: 1 or 0, default is 0, unwrap the phase or not
    10. IF_SAVE_PLOT: 1 or 0, default is 1, save the plot or not
    Output:
    1. Two figures: one for the magnitude of the RTFs and the other for the phase of the RTFs
    We save the .png and .pdf files of the figures in the save_path
    """
    # Set up latex for plotting
    plt.rcParams["text.usetex"] = True
    # Create save_path if it does not exist
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    # INPUTS
    # P_all: A list containing all RTFs to be plotted, each one is a numpy array
    # P_labels: A list containing the labels for each RTF
    # check if the length of P_all and P_labels are the same
    if len(P_all) != len(P_labels):
        raise ValueError("The number of RTFs and labels must match.")
    # Now use a few colors to plot the RTFs, black, gray, lightgray
    colors = ["gray", "black", "lightgray", "red"]
    linestypes = ["-", "-", "-.", "-"]
    linewidths = [6, 3, 3, 3]
    # Initialize the SPL arrays for each RTF
    plot_mag_all = []
    for i in range(len(P_all)):
        plot_mag_all.append(np.zeros_like(P_all[i], dtype="float"))

    # dB or linear scale
    P0 = 20 * 10 ** (-6)  # reference pressure
    if PLOT_SCALE == "linear":
        # calculate the magnitude of each RTF
        for i in range(len(P_all)):
            plot_mag_all[i] = np.abs(P_all[i])

    elif PLOT_SCALE == "dB":
        # calculate the SPL of each RTF
        for i in range(len(P_all)):
            p_rms_all = np.abs(np.sqrt(0.5 * P_all[i] * P_all[i].conjugate()))
            plot_mag_all[i] = 20 * np.log10(p_rms_all / P0)

    # Plot magnitudes of the room transfer functions
    fig = plt.figure(figsize=(18, 8))
    ax = fig.add_subplot(1, 1, 1)
    # plot the magnitude or SPL of each RTF
    for i in range(len(P_all)):
        ax.plot(
            P_freqs[i],
            plot_mag_all[i],
            label=P_labels[i],
            color=colors[i],
            linestyle=linestypes[i],
            linewidth=linewidths[i],
        )
    ax.set_xlim([P_freqs[i][0], P_freqs[i][-1]])
    ax.set_xlabel(r"\bf{Frequency (Hz)}")
    if IF_SAME_MAGSCALE == 1:
        ax.set_ylim([0, 60])
    if PLOT_SCALE == "linear":
        ax.set_ylabel(r"\bf{Magnitude}")
    elif PLOT_SCALE == "dB":
        ax.set_ylabel(r"\bf{Magnitude (dB)}")
    ax.xaxis.label.set_size(40)
    ax.yaxis.label.set_size(40)
    ax.xaxis.set_tick_params(labelsize=50)
    ax.yaxis.set_tick_params(labelsize=50)
    if IF_FREQS_Log == 1:
        ax.set_xscale("log")
    # ax.set_title('Magnitude of Transfer Functions')
    ax.legend(fontsize=35)
    # grid should apprear for every tick of the axis
    plt.grid(axis="both", which="both", linestyle=":")
    fig.tight_layout()
    fig_name = "{}/RTF_mags_{}.png".format(save_path, figure_name)
    fig_name_eps = fig_name[:-3] + "eps"
    fig_name_pdf = fig_name[:-3] + "pdf"

    if IF_SAVE_PLOT == 1:
        plt.savefig(fig_name, dpi=300, bbox_inches="tight")
        # plt.savefig(fig_name_eps, bbox_inches="tight")
        plt.savefig(fig_name_pdf, dpi=300, bbox_inches="tight")
        plt.close("all")
    else:
        plt.show()

    # Plot phases of the room transfer functions
    fig = plt.figure(figsize=(18, 8))
    ax = fig.add_subplot(1, 1, 1)
    if IF_UNWRAP_PHASE == 1:  # Unwarp the phase plots
        for i in range(len(P_all)):
            ax.plot(
                P_freqs[i],
                np.unwrap(np.angle(P_all[i])) / np.pi,
                label=P_labels[i],
                color=colors[i],
                linestyle=linestypes[i],
                linewidth=linewidths[i],
            )
    else:
        for i in range(len(P_all)):
            ax.plot(
                P_freqs[i],
                np.angle(P_all[i]) / np.pi,
                label=P_labels[i],
                color=colors[i],
                linestyle=linestypes[i],
                linewidth=linewidths[i],
            )
    ax.set_xlim([P_freqs[i][0], P_freqs[i][-1]])
    ax.set_xlabel(r"\bf{Frequency (Hz)}")
    ax.set_ylabel(r"\bf{Phase} ($\pi$)")
    # ax.set_title('Phase of Transfer Functions')
    ax.xaxis.label.set_size(40)
    ax.yaxis.label.set_size(40)
    ax.xaxis.set_tick_params(labelsize=50)
    ax.yaxis.set_tick_params(labelsize=50)
    ax.legend(fontsize=35)
    if IF_FREQS_Log == 1:
        ax.set_xscale("log")
    plt.grid(axis="both", which="both", linestyle=":")
    fig.tight_layout()
    fig_name = "{}/RTF_phase_{}.png".format(save_path, figure_name)
    fig_name_eps = fig_name[:-3] + "eps"
    fig_name_pdf = fig_name[:-3] + "pdf"

    if IF_SAVE_PLOT == 1:
        plt.savefig(fig_name, dpi=300, bbox_inches="tight")
        # plt.savefig(fig_name_eps, bbox_inches="tight")
        plt.savefig(fig_name_pdf, dpi=300, bbox_inches="tight")
        plt.close("all")
    else:
        plt.show()


def plot_results_LCs(
    freqs,
    P_DEISM_LC,
    P_DEISM_LC_matrix,
    PLOT_SCALE="dB",
    IF_UNWRAP_PHASE=1,
    IF_SAME_MAGSCALE=1,
):
    """Plot the results of the simulations."""

    # Set up latex for plotting
    plt.rcParams["text.usetex"] = True

    # Initialize the SPL arrays
    plot_mag_DEISM_LC = np.zeros_like(P_DEISM_LC, dtype="float")
    plot_mag_DEISM_LC_matrix = np.zeros_like(P_DEISM_LC_matrix, dtype="float")

    # dB or linear scale
    P0 = 20 * 10 ** (-6)  # reference pressure
    if PLOT_SCALE == "linear":
        plot_mag_DEISM_LC = np.abs(P_DEISM_LC)
        plot_mag_DEISM_LC_matrix = np.abs(P_DEISM_LC_matrix)
    elif PLOT_SCALE == "dB":
        p_rms_DEISM_LC = np.abs(np.sqrt(0.5 * P_DEISM_LC * P_DEISM_LC.conjugate()))
        p_rms_DEISM_LC_matrix = np.abs(
            np.sqrt(0.5 * P_DEISM_LC_matrix * P_DEISM_LC_matrix.conjugate())
        )
        plot_mag_DEISM_LC = 20 * np.log10(p_rms_DEISM_LC / P0)
        plot_mag_DEISM_LC_matrix = 20 * np.log10(np.abs(p_rms_DEISM_LC_matrix) / P0)
    # Plot magnitudes of the room transfer functions
    fig = plt.figure(figsize=(18, 8))
    ax = fig.add_subplot(1, 1, 1)
    ax.plot(
        freqs,
        plot_mag_DEISM_LC_matrix,
        label="DEISM LC matrix",
        color="lightgray",
        linewidth=3,
    )
    ax.plot(
        freqs,
        plot_mag_DEISM_LC,
        label="DEISM LC",
        linestyle="dotted",
        color="black",
        linewidth=3,
    )
    ax.set_xlim([freqs[0], freqs[-1]])
    ax.set_xlabel(r"\bf{Frequency (Hz)}")
    if IF_SAME_MAGSCALE == 1:
        ax.set_ylim([-10, 60])
    if PLOT_SCALE == "linear":
        ax.set_ylabel(r"\bf{Magnitude}")
    elif PLOT_SCALE == "dB":
        ax.set_ylabel(r"\bf{Magnitude (dB)}")
    ax.xaxis.label.set_size(25)
    ax.yaxis.label.set_size(25)
    ax.xaxis.set_tick_params(labelsize=25)
    ax.yaxis.set_tick_params(labelsize=25)
    # ax.set_title('Magnitude of Transfer Functions')
    ax.legend(fontsize=20)
    plt.grid(axis="both", linestyle=":")
    fig.tight_layout()
    plt.show()

    # Plot phases of the room transfer functions
    fig = plt.figure(figsize=(18, 8))
    ax = fig.add_subplot(1, 1, 1)
    if IF_UNWRAP_PHASE == 1:  # Unwarp the phase plots
        ax.plot(
            freqs,
            np.unwrap(np.angle(P_DEISM_LC_matrix)) / np.pi,
            label="DEISM LC matrix",
            color="lightgray",
            linewidth=3,
        )
        ax.plot(
            freqs,
            np.unwrap(np.angle(P_DEISM_LC)) / np.pi,
            label="DEISM LC",
            linestyle="dotted",
            color="black",
            linewidth=3,
        )
    else:
        ax.plot(
            freqs,
            np.angle(P_DEISM_LC_matrix) / np.pi,
            label="DEISM LC matrix",
            color="lightgray",
            linewidth=3,
        )
        ax.plot(
            freqs,
            np.angle(P_DEISM_LC) / np.pi,
            label="DEISM LC",
            linestyle="dotted",
            color="black",
            linewidth=3,
        )
    ax.set_xlim([freqs[0], freqs[-1]])
    ax.set_xlabel(r"\bf{Frequency (Hz)}")
    ax.set_ylabel(r"\bf{Phase} ($\pi$)")
    # ax.set_title('Phase of Transfer Functions')
    ax.xaxis.label.set_size(25)
    ax.yaxis.label.set_size(25)
    ax.xaxis.set_tick_params(labelsize=25)
    ax.yaxis.set_tick_params(labelsize=25)
    ax.legend(fontsize=20)
    plt.grid(axis="both", linestyle=":")
    fig.tight_layout()
    plt.show()
